# pylint: disable=C0401
# flake8: noqa
from continuum.metrics.logger import Logger
from continuum.metrics.metrics import (
    accuracy, accuracy_A, backward_transfer, positive_backward_transfer,
    remembering, forward_transfer, forgetting, get_model_size
)

__all__ = [
    "Logger",
    "accuracy",
    "accuracy_A",
    "backward_transfer",
    "positive_backward_transfer",
    "remembering",
    "forward_transfer",
    "forgetting",
    'get_model_size'
]



import numpy as np
import pytest
import torch
import random
import string

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental
from continuum.generators import TaskOrderGenerator, ClassOrderGenerator, HashGenerator


def gen_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


def test_task_order_generator():
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)

    scenario_generator = TaskOrderGenerator(scenario)
    sample_scenario = scenario_generator.sample()

    assert sample_scenario.nb_tasks == scenario.nb_tasks


def test_class_order_generator():
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)

    scenario_generator = ClassOrderGenerator(scenario)
    sample_scenario = scenario_generator.sample()

    assert sample_scenario.nb_tasks == scenario.nb_tasks
    assert sample_scenario.nb_classes == scenario.nb_classes
    assert (sample_scenario.classes == scenario.classes).all()


@pytest.mark.parametrize("seed",
                         [0, 41, 1992]
                         )
def test_class_order_generator(seed):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)

    scenario_generator = ClassOrderGenerator(scenario)
    sample_scenario = scenario_generator.sample(seed)
    class_order = scenario_generator.get_class_order(seed)

    assert (np.array(class_order) == np.array(sample_scenario.class_order)).all()


@pytest.mark.parametrize("seeds", [
    [0, 1],
    [1664, 41],
    [1792, 1992]
])
def test_task_order_generator_seed(seeds):
    train, test = gen_data()
    seed_0 = seeds[0]
    seed_1 = seeds[1]
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)

    scenario_generator = TaskOrderGenerator(scenario)
    task_order_0 = scenario_generator.get_task_order(seed=seed_0)
    task_order_1 = scenario_generator.get_task_order(seed=seed_1)
    task_order_0_2 = scenario_generator.get_task_order(seed=seed_0)

    assert not torch.all(task_order_0.eq(task_order_1))
    assert torch.all(task_order_0.eq(task_order_0_2))


@pytest.mark.parametrize("nb_tasks", [
    2,
    4
])
def test_task_order_generator_nb_tasks(nb_tasks):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)
    scenario_generator = TaskOrderGenerator(scenario)
    sample_scenario = scenario_generator.sample(nb_tasks=nb_tasks)

    assert sample_scenario.nb_tasks == nb_tasks


def test_hash_generator_auto():
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    NB_TASKS = 2
    list_hash = ["AverageHash", "Phash", "PhashSimple", "DhashH", "DhashV", "Whash", "ColorHash"
                 ] # , "CropResistantHash"
    scenario_generator = HashGenerator(cl_dataset=dummy,
                                       list_hash=list_hash,
                                       nb_tasks=2,
                                       transformations=None,
                                       filename_hash_indexes=None,
                                       split_task="auto")

    sample_scenario = scenario_generator.sample()

    assert sample_scenario.nb_tasks == NB_TASKS


def test_hash_generator_auto_full():
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    list_hash = ["AverageHash", "Phash", "PhashSimple", "DhashH", "DhashV", "Whash", "ColorHash"
                 ] # , "CropResistantHash"
    scenario_generator = HashGenerator(cl_dataset=dummy,
                                       list_hash=list_hash,
                                       nb_tasks=None,
                                       transformations=None,
                                       filename_hash_indexes=None,
                                       split_task="auto")

    sample_scenario = scenario_generator.sample()
    assert sample_scenario.nb_tasks >= 2
    sample_scenario = scenario_generator.sample()
    assert sample_scenario.nb_tasks >= 2


@pytest.mark.parametrize("nb_tasks, list_hash_name", [
    (2, ["AverageHash", "Whash", "ColorHash", "DhashV"]), #"CropResistantHash"
    (4, ["DhashH", "DhashV", "Whash", "ColorHash"]),
    (3, ["AverageHash", "DhashH"]),
    (5, ["AverageHash", "Phash", "PhashSimple", "DhashH", "DhashV", "Whash", "ColorHash",
         ]) #"CropResistantHash"
])
def test_hash_generator_balanced(nb_tasks, list_hash_name):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)

    print(list_hash_name)
    scenario_generator = HashGenerator(cl_dataset=dummy,
                                       list_hash=list_hash_name,
                                       nb_tasks=nb_tasks,
                                       transformations=None,
                                       filename_hash_indexes=None,
                                       split_task="balanced")

    sample_scenario = scenario_generator.sample()
    assert sample_scenario.split_task == "balanced"
    assert sample_scenario.hash_name in list_hash_name
    # test nb of samples per task
    assert (len(sample_scenario[0]) == len(sample_scenario[1]))\
           or (len(sample_scenario[0]) + 1 == len(sample_scenario[1]))



from typing import Any

import torch
import numpy as np


def herd_random(x: np.ndarray, y: np.ndarray, t: np.ndarray, z: Any, nb_per_class: int) -> np.ndarray:
    """Herd randomly examples for rehearsal.

    :param x: Input data (images, paths, etc.)
    :param y: Labels of the data.
    :param t: Task ids of the data.
    :param z: Extra info, here unused.
    :param nb_per_class: Number of samples to herd per class.
    :return: The sampled data x, y, t.
    """
    indexes = []

    for class_id in np.unique(y):
        class_indexes = np.where(y == class_id)[0]
        indexes.append(
            np.random.choice(
                class_indexes,
                size=min(nb_per_class, len(class_indexes)),
                replace=False
            )
        )

    indexes =  np.concatenate(indexes)
    return x[indexes], y[indexes], t[indexes]


def herd_closest_to_cluster(
    x: np.ndarray,
    y: np.ndarray,
    t: np.ndarray,
    features: np.ndarray,
    nb_per_class: np.ndarray
) -> np.ndarray:
    """Herd the samples whose features is the closest to their class mean.

    :param x: Input data (images, paths, etc.)
    :param y: Labels of the data.
    :param t: Task ids of the data.
    :param features: Features of shape (nb_samples, nb_dim).
    :param nb_per_class: Number of samples to herd per class.
    :return: The sampled data x, y, t.
    """
    if len(features.shape) != 2:
        raise ValueError(f"Expected features to have 2 dimensions, not {len(features.shape)}d.")
    indexes = []

    for class_id in np.unique(y):
        class_indexes = np.where(y == class_id)[0]
        class_features = features[class_indexes]
        class_mean = np.mean(class_features, axis=1, keepdims=True)

        dist_to_mean = np.linalg.norm(class_mean - class_features, axis=1)
        tmp_indexes = dist_to_mean.argsort()[:nb_per_class]

        indexes.append(class_indexes[tmp_indexes])

    indexes =  np.concatenate(indexes)
    return x[indexes], y[indexes], t[indexes]


def herd_closest_to_barycenter(
    x: np.ndarray,
    y: np.ndarray,
    t: np.ndarray,
    features: np.ndarray,
    nb_per_class: int
) -> np.ndarray:
    """Herd the samples whose features is the closest to their moving barycenter.

    Reference:
        * iCaRL: Incremental Classifier and Representation Learning
          Sylvestre-Alvise Rebuffi, Alexander Kolesnikov, Georg Sperl, Christoph H. Lampert
          CVPR 2017

    :param x: Input data (images, paths, etc.)
    :param y: Labels of the data.
    :param t: Task ids of the data.
    :param features: Features of shape (nb_samples, nb_dim).
    :param nb_per_class: Number of samples to herd per class.
    :return: The sampled data x, y, t.
    """
    if len(features.shape) != 2:
        raise ValueError(f"Expected features to have 2 dimensions, not {len(features.shape)}d.")

    indexes = []

    for class_id in np.unique(y):
        class_indexes = np.where(y == class_id)[0]
        class_features = features[class_indexes]

        D = class_features.T
        D = D / (np.linalg.norm(D, axis=0) + 1e-8)
        mu = np.mean(D, axis=1)
        herding_matrix = np.zeros((class_features.shape[0],))

        w_t = mu
        iter_herding, iter_herding_eff = 0, 0

        while not (
            np.sum(herding_matrix != 0) == min(nb_per_class, class_features.shape[0])
        ) and iter_herding_eff < 1000:
            tmp_t = np.dot(w_t, D)
            ind_max = np.argmax(tmp_t)
            iter_herding_eff += 1
            if herding_matrix[ind_max] == 0:
                herding_matrix[ind_max] = 1 + iter_herding
                iter_herding += 1

            w_t = w_t + mu - D[:, ind_max]

        herding_matrix[np.where(herding_matrix == 0)[0]] = 10000

        tmp_indexes = herding_matrix.argsort()[:nb_per_class]
        indexes.append(class_indexes[tmp_indexes])

    indexes =  np.concatenate(indexes)
    return x[indexes], y[indexes], t[indexes]



from typing import Callable, List, Optional

import numpy as np
from torchvision import transforms

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import InstanceIncremental
from continuum.tasks import TaskSet


class TransformationIncremental(InstanceIncremental):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: Every task contains the same data with different transformations.
    It is a cheap way to create instance incremental scenarios.
    Moreover, it is easier to analyse what algorithms forget or not.
    Classic transformation incremental scenarios are "permutations" and "rotations".

    :param cl_dataset: A continual dataset.
    :param incremental_transformations: list of transformations to apply to specific tasks
    :param base_transformations: List of transformation to apply to all tasks.
    :param shared_label_space: If true same data with different transformation have same label
    """

    def __init__(
            self,
            cl_dataset: _ContinuumDataset,
            incremental_transformations: List[List[Callable]],
            base_transformations: List[Callable] = None,
            shared_label_space=True
    ):
        nb_tasks = len(incremental_transformations)
        if incremental_transformations is None:
            raise ValueError("For this scenario a list transformation should be set")

        self.inc_trsf = incremental_transformations
        #self._nb_tasks = self._setup(nb_tasks)
        self.shared_label_space = shared_label_space

        super().__init__(
            cl_dataset=cl_dataset, nb_tasks=nb_tasks, transformations=base_transformations
        )

        self.num_classes_per_task = len(np.unique(self.dataset[1]))  # the num of classes is the same for all task is this scenario

    @property
    def nb_classes(self) -> int:
        """Total number of classes in the whole continual setting."""
        if self.shared_label_space:
            nb_classes = len(np.unique(self.dataset[1]))
        else:
            nb_classes = len(np.unique(self.dataset[1])) * self._nb_tasks
        return nb_classes

    def get_task_transformation(self, task_index):
        return transforms.Compose(self.inc_trsf[task_index] + self.trsf.transforms)

    def update_task_indexes(self, task_index):
        new_t = np.ones(len(self.dataset[1])) * task_index
        self.dataset = (self.dataset[0], self.dataset[1], new_t)

    def update_labels(self, task_index):
        # wrong
        # new_y = self.dataset[1] + task_index * self.num_classes_per_task
        # we update incrementally then update is simply:
        if task_index > 0:
            new_y = self.dataset[1] + self.num_classes_per_task
            self.dataset = (self.dataset[0], new_y, self.dataset[2])

    def __getitem__(self, task_index):
        """Returns a task by its unique index.

        :param task_index: The unique index of a task, between 0 and len(loader) - 1. Or it could
                           be a list or a numpy array or even a slice.
        :return: A train PyTorch's Datasets.
        """
        x, y, _ = self.dataset

        if isinstance(task_index, slice):
            # Convert a slice to a list and respect the Python's advanced indexing conventions
            start = task_index.start if task_index.start is not None else 0
            stop = task_index.stop if task_index.stop is not None else len(self) + 1
            step = task_index.step if task_index.step is not None else 1
            task_index = list(range(start, stop, step))
            if len(task_index) == 0:
                raise ValueError(f"Invalid slicing resulting in no data (start={start}, end={stop}, step={step}).")
        elif isinstance(task_index, np.ndarray):
            task_index = list(task_index)
        elif isinstance(task_index, int):
            task_index = [task_index]
        else:
            raise TypeError(f"Invalid type of task index {type(task_index).__name__}.")

        task_index = set([_handle_negative_indexes(ti, len(self)) for ti in task_index])

        t = np.concatenate([
            (np.ones(len(x)) * ti).astype(np.int32) for ti in task_index
        ])
        x = np.concatenate([
            x for _ in range(len(task_index))
        ])
        if self.shared_label_space:
            y = np.concatenate([
                y for _ in range(len(task_index))
            ])
        else:
            # Different transformations have different labels even though
            # the original images were the same
            y = np.concatenate([
                y + ti * self.num_classes_per_task for ti in task_index
            ])

        trsf = [  # Non-used tasks have a None trsf
            self.get_task_transformation(ti)
            if ti in task_index else None
            for ti in range(len(self))
        ]

        return TaskSet(x, y, t, trsf, data_type=self.cl_dataset.data_type)


def _handle_negative_indexes(index: int, total_len: int) -> int:
    while index < 0:
        index += total_len
    return index



import numpy as np

from continuum.datasets import MNIST


class RainbowMNIST(MNIST):
    def __init__(
        self,
        *args,
        color: str = "red",
        rotation: float = 0.,
        scale=1.,
        **kwargs
    ):
        if rotation != 0. or scale != 1.:
            raise NotImplementedError(
                "Rotation and scale are not supported yet for RainbowMNIST."
            )
        if color not in ("red", "blue", "green"):
            raise ValueError(
                f"Only red, blue, and green colors are available, not {color}."
            )

        super().__init__(*args, **kwargs)

        self.color = color
        self.rotation = rotation
        self.scale = scale

    def get_data(self):
        x, y, t = super().get_data()
        x = x[..., None]

        if self.color == "red":
            c = np.zeros((x.shape[0], x.shape[1], x.shape[2], 2), dtype=np.uint8)
            x = np.concatenate((x, c), axis=-1)
        elif self.color == "green":
            c = np.zeros((x.shape[0], x.shape[1], x.shape[2], 1), dtype=np.uint8)
            x = np.concatenate((c, x, c), axis=-1)
        else:
            c = np.zeros((x.shape[0], x.shape[1], x.shape[2], 2), dtype=np.uint8)
            x = np.concatenate((c, x), axis=-1)

        return x, y, t



import abc

import torch
import numpy as np

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import _BaseScenario, create_subscenario, ClassIncremental, HashedScenario


class _BaseGenerator(abc.ABC):
    """Abstract loader.

    DO NOT INSTANTIATE THIS CLASS.

    :param scenario: A Continuum scenarios
    """

    def __init__(
            self,
            scenario: _BaseScenario
    ) -> None:
        self.base_scenario = scenario
        self.nb_generator = torch.Generator()

    @abc.abstractmethod
    def sample(self, seed: int = None, nb_tasks: int = None) -> int:
        """"method to sample a scenario from the generator."""
        raise NotImplementedError


class TaskOrderGenerator(_BaseGenerator):
    """Task Order Generator, generate sub-scenario from a base scenario simply by changing task order.

        :param scenario: the base scenario to use to generate sub-scenarios
        """

    def __init__(
            self,
            scenario: _BaseScenario
    ) -> None:
        super().__init__(scenario)

        self.base_scenario = scenario
        self.nb_generator = torch.Generator()
        self.task_order = None

    def get_task_order(self, seed, nb_tasks: int = None):
        self.nb_generator.manual_seed(seed)

        # generate a random task order
        task_order = torch.randperm(self.base_scenario.nb_tasks, generator=self.nb_generator)
        if nb_tasks is None:
            nb_tasks = self.base_scenario.nb_tasks
        return task_order[:nb_tasks]

    def sample(self, seed: int = None, nb_tasks: int = None) -> _BaseScenario:
        # seed the generator
        if seed is None:
            seed = np.random.randint(10000)

        # generate a random task order
        task_order = self.get_task_order(seed, nb_tasks)

        subscenario = create_subscenario(self.base_scenario, task_order[:nb_tasks])

        return subscenario


class ClassOrderGenerator(_BaseGenerator):
    """Class Order Generator, generate sub-scenario from a base scenario simply by changing class order.
    The difference with TaskOrderGenerator is that the classes inside a same task change.
    This class is only compatible with ClassIncremental scenarios.

        :param scenario: the base scenario to use to generate sub-scenarios
        """

    def __init__(
            self,
            scenario: ClassIncremental
    ) -> None:
        super().__init__(scenario)

        self.base_scenario = scenario
        self.list_classes = scenario.classes
        self.nb_generator = torch.Generator()
        self.task_order = None

    def get_class_order(self, seed):
        self.nb_generator.manual_seed(seed)
        # generate a random task order
        class_order = torch.randperm(len(self.list_classes), generator=self.nb_generator)
        return class_order

    def sample(self, seed: int = None, nb_tasks: int = None) -> _BaseScenario:

        if nb_tasks is None:
            nb_tasks = self.base_scenario.nb_tasks

        if nb_tasks != self.base_scenario.nb_tasks:
            AssertionError("You can not change the number of tasks in the generator")

        # seed the generator
        if seed is None:
            seed = np.random.randint(10000)

        # generate a random class order
        class_order = self.get_class_order(seed)
        new_list_class = self.list_classes[class_order]

        # We create a scenario from base_scenario
        initial_increment = self.base_scenario.initial_increment
        nb_tasks = self.base_scenario.nb_tasks
        increment = self.base_scenario.increment
        cl_dataset = self.base_scenario.cl_dataset
        transformations = self.base_scenario.transformations

        scenario = ClassIncremental(cl_dataset=cl_dataset,
                                    nb_tasks=nb_tasks,
                                    increment=increment,
                                    initial_increment=initial_increment,
                                    transformations=transformations,
                                    class_order=new_list_class)

        return scenario


class HashGenerator(_BaseGenerator):
    """Hash Generator, generate scenario from a set of parameters similar to scenario parameters and a list of hash name
    Hash generator will create a HashedScenario from a hash name randomly sampled in the list.

    :param cl_dataset: A continual dataset.
    :param list_hash: list of hash name
    :param nb_tasks: nb_tasks of each scenario, if None it will be automatically set
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    :param filename_hash_indexes: base name of a file to save scenarios indexes and reload them after
    :param split_task: Define if the task split will be automatic by clusterization of hashes or manually balanced
    """

    def __init__(
            self,
            cl_dataset: _ContinuumDataset,
            list_hash=None,
            nb_tasks=None,
            transformations=None,
            filename_hash_indexes=None,
            split_task="auto"
    ) -> None:
        self.cl_dataset = cl_dataset
        self.nb_tasks = nb_tasks
        self.transformations = transformations
        self.filename_hash_indexes = filename_hash_indexes
        self.split_task = split_task
        self._hash_name = None

        self.all_hashs = ["AverageHash", "Phash", "PhashSimple", "DhashH", "DhashV", "Whash", "ColorHash"
                          ] # , "CropResistantHash"

        # create default scenario to test parameters
        self.base_scenario = HashedScenario(cl_dataset=self.cl_dataset,
                                            hash_name="AverageHash",
                                            nb_tasks=nb_tasks,
                                            transformations=transformations,
                                            filename_hash_indexes=f"{filename_hash_indexes}_AverageHash",
                                            split_task=self.split_task)
        if list_hash is None:
            self.list_hash = self.all_hashs
        else:
            # all items of list_hash should be in self.all_hashs
            if not all(item in self.all_hashs for item in list_hash):
                AssertionError("Unknown hash name")
            else:
                self.list_hash = list_hash
        super().__init__(self.base_scenario)

    def get_rand_hash_name(self, seed):
        self.nb_generator.manual_seed(seed)
        # generate a random task order
        hash_id = torch.randperm(len(self.list_hash), generator=self.nb_generator)[0]
        return self.list_hash[hash_id]

    def sample(self, seed: int = None, nb_tasks: int = None) -> _BaseScenario:
        ''''create one scenario with a ramdomly sampled hash_name from self.list_hash'''

        if nb_tasks is None and self.nb_tasks is not None:
            nb_tasks = self.nb_tasks

        # seed the generator
        if seed is None:
            seed = np.random.randint(10000)

        # generate a random class order
        self._hash_name = self.get_rand_hash_name(seed)

        # We create a scenario from base_scenario
        scenario = HashedScenario(cl_dataset=self.cl_dataset,
                                  hash_name=self._hash_name,
                                  nb_tasks=nb_tasks,
                                  transformations=self.transformations,
                                  filename_hash_indexes=f"{self.filename_hash_indexes}_{self._hash_name}",
                                  split_task=self.split_task)

        return scenario

    @property
    def hash_name(self):
        ''''Hash name of the previous sampled scenario'''
        return self._hash_name



import numpy as np
import pytest
import os

from continuum.datasets import CIFAR10, CIFAR100, TinyImageNet200, InMemoryDataset
from continuum.scenarios import HashedScenario

DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")


@pytest.mark.slow
@pytest.mark.parametrize("hash_name",
                         ["Whash",
                          "DhashV",
                          "DhashH",
                          "PhashSimple",
                          "Phash",
                          "AverageHash",
                          "ColorHash"])  # , "CropResistantHash"]) # too long CropResistantHash
@pytest.mark.parametrize(("dataset", "shape", "split_task"),
                         [(CIFAR10, [32, 32, 3], "auto"),
                          (CIFAR100, [32, 32, 3], "auto"),
                          (TinyImageNet200, [64, 64, 3], "auto"),
                          (CIFAR10, [32, 32, 3], "balanced"),
                          (CIFAR100, [32, 32, 3], "balanced"),
                          (TinyImageNet200, [64, 64, 3], "balanced")])
def test_visualization_HashedScenario(hash_name, dataset, shape, split_task):
    if split_task == "balanced":
        num_tasks = 5
    else:
        num_tasks = None

    dataset = dataset(data_path=DATA_PATH, download=False, train=True)
    scenario = HashedScenario(cl_dataset=dataset,
                              hash_name=hash_name,
                              nb_tasks=num_tasks,
                              split_task=split_task)

    assert scenario.nb_tasks > 1

    folder = "tests/samples/hashed_scenario/"
    if not os.path.exists(folder):
        os.makedirs(folder)

    # test default parameters
    for task_id, taskset in enumerate(scenario):
        taskset.plot(path=folder,
                     title="{}_HashedScenario_{}_{}_{}.jpg".format(split_task,
                                                                   type(dataset).__name__,
                                                                   hash_name,
                                                                   task_id),
                     nb_samples=100,
                     shape=shape)


def numpy_data():
    x_train = []
    y_train = []
    for i in range(10):
        x_train.append(np.random.randn(5, 10, 10, 3))
        y_train.append(np.ones(5) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    return (x_train, y_train.astype(int))


@pytest.mark.parametrize("hash_name",
                         ["Whash",
                          "DhashV",
                          "DhashH",
                          "PhashSimple",
                          "Phash",
                          "AverageHash",
                          "ColorHash"])  # , "CropResistantHash"
def test_HashedScenario_save_indexes(tmpdir, hash_name):
    num_tasks = 2
    x, y = numpy_data()
    dataset = InMemoryDataset(x, y, None, data_type="image_array")

    filename_indexes = os.path.join(tmpdir, f"{hash_name}.npy")
    if os.path.exists(filename_indexes):
        os.remove(filename_indexes)

    if os.path.exists(filename_indexes):
        AssertionError(f"{filename_indexes} should have been delete.")

    # test save the indexes array
    scenario = HashedScenario(cl_dataset=dataset,
                              hash_name=hash_name,
                              nb_tasks=num_tasks,
                              filename_hash_indexes=filename_indexes)

    # test load the indexes array
    scenario = HashedScenario(cl_dataset=dataset,
                              hash_name=hash_name,
                              nb_tasks=num_tasks,
                              filename_hash_indexes=filename_indexes)

    # delete test indexes
    os.remove(filename_indexes)


@pytest.mark.parametrize("hash_name",
                         ["Whash",
                          "DhashV",
                          "DhashH",
                          "PhashSimple",
                          "Phash",
                          "AverageHash",
                          "ColorHash"])  # , "CropResistantHash"
def test_HashedScenario_automatic_task_number(hash_name):
    x, y = numpy_data()
    dataset = InMemoryDataset(x, y, None, data_type="image_array")

    # test when nb_tasks is set to None
    scenario = HashedScenario(cl_dataset=dataset,
                              hash_name=hash_name,
                              nb_tasks=None,
                              split_task="auto")

    if scenario.nb_tasks is None or scenario.nb_tasks < 2:
        AssertionError("nb_tasks should have been set automatically to more than one")



from typing import Callable, List, Tuple, Union

from torchvision import transforms

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import TransformationIncremental


class Rotations(TransformationIncremental):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: Rotations scenario is a new instance scenario.
              For each task data is rotated from a certain angle.

    :param cl_dataset: A continual dataset.
    :param nb_tasks: The scenario's number of tasks.
    :param list_degrees: list of rotation in degree (int) or list of range. e.g. (0, (40,45), 90).
    :param base_transformations: Preprocessing transformation to applied to data before rotation.
    :param shared_label_space: If true same data with different transformation have same label
    """

    def __init__(
        self,
        cl_dataset: _ContinuumDataset,
        list_degrees: Union[List[Tuple], List[int]],
        nb_tasks: Union[int, None] = None,
        base_transformations: List[Callable] = None,
        shared_label_space=True
    ):

        if nb_tasks is not None and len(list_degrees) != nb_tasks:
            raise ValueError(
                f"The nb of tasks ({nb_tasks}) != number of angles "
                f"tuples ({len(list_degrees)}) set in the list"
            )

        trsfs = self._generate_transformations(list_degrees)

        super().__init__(
            cl_dataset=cl_dataset,
            incremental_transformations=trsfs,
            base_transformations=base_transformations,
            shared_label_space=shared_label_space
        )

    def _generate_transformations(self, degrees):
        trsfs = []
        min_deg, max_deg = None, None

        for deg in degrees:
            if isinstance(deg, int):
                min_deg, max_deg = deg, deg
            elif len(deg) == 2:
                min_deg, max_deg = deg
            else:
                raise ValueError(
                    f"Invalid list of degrees ({degrees}). "
                    "It should contain either integers (-deg, +deg) or "
                    "tuples (range) of integers (deg_a, deg_b)."
                )

            trsfs.append([transforms.RandomAffine(degrees=[min_deg, max_deg])])

        return trsfs



"""Continuum lib.

Made by Arthur Douillard and Timothee Lesort.

The goal of this library is to provide clean and simple to use utilities for
Continual Learning.
"""
# pylint: disable=C0401
# flake8: noqa
from continuum import datasets
from continuum.scenarios import *
from continuum.tasks import *
from continuum.metrics import *
from continuum.viz import plot_samples
from continuum import generators


__version__ = "1.1.1"



# pylint: disable=C0401
# flake8: noqa
from continuum.scenarios.base import _BaseScenario
from continuum.scenarios.continual_scenario import ContinualScenario
from continuum.scenarios.class_incremental import ClassIncremental
from continuum.scenarios.instance_incremental import InstanceIncremental
from continuum.scenarios.transformation_incremental import TransformationIncremental
from continuum.scenarios.rotations import Rotations
from continuum.scenarios.permutations import Permutations
from continuum.scenarios.segmentation import SegmentationClassIncremental
from continuum.scenarios.hashed import HashedScenario

from continuum.scenarios.scenario_utils import create_subscenario

__all__ = [
    "ContinualScenario",
    "ClassIncremental",
    "InstanceIncremental",
    "Rotations",
    "Permutations",
    "TransformationIncremental",
    "SegmentationClassIncremental",
    "HashedScenario"
]



import os

import pytest
from torchvision.transforms import Resize, ToTensor

from continuum.datasets import CUB200
from continuum.scenarios import ClassIncremental

DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")

'''
Test the visualization with instance_class scenario
'''
@pytest.mark.slow
def test_scenario_CUB200_ClassIncremental():
    dataset = CUB200(DATA_PATH, train=True, transform=None)
    scenario = ClassIncremental(dataset, increment=100, transformations=[Resize((224, 224)), ToTensor()])

    print(f"Nb classes : {scenario.nb_classes} ")
    print(f"Nb tasks : {scenario.nb_tasks} ")
    for task_id, task_set in enumerate(scenario):
        print(f"Task {task_id} : {task_set.nb_classes} classes")
        task_set.plot(path="Archives/Samples/CUB200/CI",
                      title="CUB200_ClassIncremental_{}.jpg".format(task_id),
                      nb_samples=100)



import os
from typing import Iterable, Set, Tuple, Union

import numpy as np

from continuum import download
from continuum.datasets.base import _SemanticSegmentationDataset
from continuum.transforms.segmentation import ToTensor


class PascalVOC2012(_SemanticSegmentationDataset):
    """PascalVOC2012 Semantic Segmentation Dataset.

    :param data_path: Path where the data is present or will be downloaded.
    :param download: Whether to download.
    :param augmented: Use augmented version of PascalVOC with more images (recommended).
    """

    data_url = "http://pjreddie.com/media/files/VOCtrainval_11-May-2012.tar"
    segmentation_url = "http://cs.jhu.edu/~cxliu/data/SegmentationClassAug.zip"
    split_url = "http://cs.jhu.edu/~cxliu/data/list.zip"

    def __init__(self, data_path: str = "", train: bool = True, download: bool = True, augmented: bool = True) -> None:
        super().__init__(data_path=data_path, train=train, download=download)
        self.augmented = augmented

    def _download(self):
        # Downloading images
        if not os.path.exists(os.path.join(self.data_path, "VOCdevkit")):
            path = os.path.join(self.data_path, "VOCtrainval_11-May-2012.tar")
            if not os.path.exists(path):
                print("Downloading Pascal VOC segmentation maps...")
                download.download(self.data_url, self.data_path)
            print("Uncompressing images...")
            download.untar(path)

        # Downloading segmentation maps
        if not os.path.exists(os.path.join(self.data_path, "SegmentationClassAug")):
            path = os.path.join(self.data_path, "SegmentationClassAug.zip")
            if not os.path.exists(path):
                print("Downloading Pascal VOC segmentation maps...")
                download.download(self.segmentation_url, self.data_path)
            print("Uncompressing segmentation maps...")
            download.unzip(path)

        # Downloading train/val/test indexes
        if not os.path.exists(os.path.join(self.data_path, "list")):
            path = os.path.join(self.data_path, "list.zip")
            if not os.path.exists(path):
                print("Downloading Pascal VOC train/val/test indexes...")
                download.download(self.split_url, self.data_path)
            print("Uncompressing train/val/test indexes...")
            download.unzip(path)

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.train and self.augmented:
            list_path = os.path.join(self.data_path, "list", "train_aug.txt")
        elif self.train:
            list_path = os.path.join(self.data_path, "list", "train.txt")
        else:
            list_path = os.path.join(self.data_path, "list", "val.txt")

        image_paths, map_paths = [], []
        with open(list_path, "r") as f:
            for line in f:
                p1, p2 = line.split(" ")
                image_paths.append(os.path.join(self.data_path, "VOCdevkit", "VOC2012", p1[1:].strip()))
                map_paths.append(os.path.join(self.data_path, p2[1:].strip()))

        return np.array(image_paths), np.array(map_paths), None



from copy import copy
from typing import Tuple, Union, Optional, List

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset as TorchDataset
from torchvision import transforms

from continuum.viz import plot_samples


class TaskSet(TorchDataset):
    """A task dataset returned by the CLLoader.

    :param x: The data, either image-arrays or paths to images saved on disk.
    :param y: The targets, not one-hot encoded.
    :param t: The task id of each sample.
    :param trsf: The transformations to apply on the images.
    :param data_type: Type of the data, either "image_path", "image_array",
                      "text", "tensor" or "segmentation".
    """

    def __init__(
            self,
            x: np.ndarray,
            y: np.ndarray,
            t: np.ndarray,
            trsf: Union[transforms.Compose, List[transforms.Compose]],
            target_trsf: Optional[Union[transforms.Compose, List[transforms.Compose]]] = None,
            data_type: str = "image_array",
            bounding_boxes: Optional[np.ndarray] = None
    ):
        self._x, self._y, self._t = x, y, t

        # if task index are not provided t is always -1
        if self._t is None:
            self._t = -1 * np.ones_like(y)

        self.trsf = trsf
        self.target_trsf = target_trsf
        self.data_type = data_type
        self.bounding_boxes = bounding_boxes

        self._to_tensor = transforms.ToTensor()

    @property
    def nb_classes(self):
        """The number of classes contained in the current task."""
        return len(np.unique(self._y))

    def get_classes(self):
        """Array of all classes contained in the current task."""
        return np.unique(self._y)

    def concat(self, *task_sets):
        """Concat others task sets.

        :param task_sets: One or many task sets.
        """
        for task_set in task_sets:
            if task_set.data_type != self.data_type:
                raise Exception(
                    f"Invalid data type {task_set.data_type} != {self.data_type}"
                )
            self.add_samples(task_set._x, task_set._y, task_set._t)

    def add_samples(self, x: np.ndarray, y: np.ndarray, t: Union[None, np.ndarray] = None):
        """Add memory for rehearsal.

        :param x: Sampled data chosen for rehearsal.
        :param y: The associated targets of `x_memory`.
        :param t: The associated task ids. If not provided, they will be
                         defaulted to -1.
        """
        self._x = np.concatenate((self._x, x))
        self._y = np.concatenate((self._y, y))
        if t is not None:
            self._t = np.concatenate((self._t, t))
        else:
            self._t = np.concatenate((self._t, -1 * np.ones(len(x))))

    def plot(
            self,
            path: Union[str, None] = None,
            title: str = "",
            nb_samples: int = 100,
            shape: Optional[Tuple[int, int]] = None,
    ) -> None:
        """Plot samples of the current task, useful to check if everything is ok.

        :param path: If not None, save on disk at this path.
        :param title: The title of the figure.
        :param nb_samples: Amount of samples randomly selected.
        :param shape: Shape to resize the image before plotting.
        """
        plot_samples(self, title=title, path=path, nb_samples=nb_samples,
                     shape=shape, data_type=self.data_type)

    def __len__(self) -> int:
        """The amount of images in the current task."""
        return self._x.shape[0]

    def get_random_samples(self, nb_samples):
        nb_tot_samples = self._x.shape[0]
        indexes = np.random.randint(0, nb_tot_samples, nb_samples)
        return self.get_samples(indexes)

    def get_samples(self, indexes):
        samples, targets, tasks = [], [], []

        w, h = None, None
        for index in indexes:
            # we need to use __getitem__ to have the transform used
            sample, y, t = self[index]

            if self.data_type in ["image_path", "image_array"]:
                # we check dimension of images
                if w is None:
                    w, h = sample.shape[:2]
                elif w != sample.shape[0] or h != sample.shape[1]:
                    raise Exception(
                        "Images dimension are inconsistent, resize them to a "
                        "common size using a transformation.\n"
                        "For example, give to the scenario you're using as `transformations` argument "
                        "the following: [transforms.Resize((224, 224)), transforms.ToTensor()]"
                    )

            samples.append(sample)
            targets.append(y)
            tasks.append(t)

        return _tensorize_list(samples), _tensorize_list(targets), _tensorize_list(tasks)

    def get_raw_samples(self, indexes=None):
        """Get samples without preprocessing, for split train/val for example."""
        if indexes is None:
            return self._x, self._y, self._t
        return self._x[indexes], self._y[indexes], self._t[indexes]

    def get_sample(self, index: int) -> np.ndarray:
        """Returns a Pillow image corresponding to the given `index`.

        :param index: Index to query the image.
        :return: A Pillow image.
        """
        x = self._x[index]

        if self.data_type in ("image_path", "segmentation"):
            x = Image.open(x).convert("RGB")
        elif self.data_type == "image_array":
            x = Image.fromarray(x.astype("uint8"))
        elif self.data_type == "tensor":
            if not torch.is_tensor(x):
                x = torch.tensor(x)
        elif self.data_type == "text":
            pass

        return x

    def __getitem__(self, index: int) -> Tuple[np.ndarray, int, int]:
        """Method used by PyTorch's DataLoaders to query a sample and its target."""
        x = self.get_sample(index)
        y = self._y[index]
        t = self._t[index]

        if self.bounding_boxes is not None:
            bbox = self.bounding_boxes[index]
            x = x.crop((
                max(bbox[0], 0),  # x1
                max(bbox[1], 0),  # y1
                min(bbox[2], x.size[0]),  # x2
                min(bbox[3], x.size[1]),  # y2
            ))

        if self.data_type == "text":
            x, y, t = self._prepare_text(x, y, t)
        elif self.data_type == "segmentation":
            x, y, t = self._prepare_segmentation(x, y, t)
        elif self.data_type == "image_array" or self.data_type == "image_path":
            x, y, t = self._prepare(x, y, t)
        else:  # self.data_type == "tensor"
            pass

        if self.target_trsf is not None:
            y = self.get_task_target_trsf(t)(y)

        return x, y, t

    def get_task_trsf(self, t: int):
        if isinstance(self.trsf, list):
            return self.trsf[t]
        return self.trsf

    def get_task_target_trsf(self, t: int):
        if isinstance(self.target_trsf, list):
            return self.target_trsf[t]
        return self.target_trsf

    def _prepare(self, x, y, t):
        if self.trsf is not None:
            x = self.get_task_trsf(t)(x)
        if not isinstance(x, torch.Tensor):
            x = self._to_tensor(x)

        return x, y, t

    def _prepare_segmentation(self, x, y, t):
        y = Image.open(y)
        if self.trsf is not None:
            x, y = self.get_task_trsf(t)(x, y)

        if not isinstance(x, torch.Tensor):
            x = self._to_tensor(x)
        if not isinstance(y, torch.Tensor):
            y = self._to_tensor(y)

        return x, y, t

    def _prepare_text(self, x, y, t):
        # Nothing in particular for now, TODO latter
        return x, y, t


def _tensorize_list(x):
    if isinstance(x[0], torch.Tensor):
        return torch.stack(x)
    return torch.tensor(x)



import os

import pytest
import numpy as np
from torchvision.transforms import Resize, ToTensor

from continuum.datasets import Stream51
from continuum.scenarios import InstanceIncremental
from continuum.scenarios import ClassIncremental


DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")


'''
Test the visualization with instance_class scenario
'''
@pytest.mark.slow
def test_scenario_clip_ClassIncremental(tmpdir):

    dataset = Stream51(DATA_PATH, task_criterion="clip")
    # dataset = Stream51('../Datasets', task_criterion="video")
    # scenario = InstanceIncremental(dataset, transformations=[Resize((224, 224)), ToTensor()])
    scenario = ClassIncremental(dataset, increment=1, transformations=[Resize((224, 224)), ToTensor()])

    print(f"Nb classes : {scenario.nb_classes} ")
    print(f"Nb tasks : {scenario.nb_tasks} ")
    for task_id, task_set in enumerate(scenario):
        print(f"Task {task_id} : {task_set.nb_classes} classes")
        task_set.plot(path="tests/samples/stream51/ci",
                      title="Stream51_InstanceIncremental_{}.jpg".format(task_id),
                      nb_samples=100)



'''
Test the visualization with instance scenario
'''
@pytest.mark.slow
def test_scenario_clip_InstanceIncremental(tmpdir):

    dataset = Stream51(DATA_PATH, task_criterion="clip")
    scenario = InstanceIncremental(dataset, transformations=[Resize((224, 224)), ToTensor()])

    print(f"Nb classes : {scenario.nb_classes} ")
    print(f"Nb tasks : {scenario.nb_tasks} ")
    for task_id, task_set in enumerate(scenario):
        print(f"Task {task_id} : {task_set.nb_classes} classes")
        task_set.plot(path="tests/samples/stream51/ci",
                      title="Stream51_InstanceIncremental_{}.jpg".format(task_id),
                      nb_samples=100)



import os
from multiprocessing import Pool, cpu_count
from typing import Callable, List, Union, Optional

import imagehash
import numpy as np
from PIL import Image
from sklearn.metrics import pairwise_distances
from sklearn.cluster import KMeans, MeanShift
from sklearn.decomposition import PCA
from scipy.spatial.distance import hamming

from continuum.datasets import InMemoryDataset
from continuum.datasets import _ContinuumDataset
from continuum.scenarios import ContinualScenario


def sort_hash(list_hash):
    size = len(list_hash)

    # multithread hash evaluation without changing list order
    with Pool(min(8, cpu_count())) as p:
        list_hash_norm = p.map(np.linalg.norm, list_hash)

    assert len(list_hash_norm)==size

    return sorted(range(size), key=lambda k: list_hash_norm[k])


def get_array_list(list_bin_str_hash):
    list_array = []
    for str_hash in list_bin_str_hash:
        array = np.array([int(hex_str, 16) for hex_str in str_hash])
        list_array.append(array)

    return list_array


class HashedScenario(ContinualScenario):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: the scenario is entirely defined by the task label vector in the cl_dataset

    :param cl_dataset: A continual dataset.
    :param hash_name: Name of the hash function that will be used to create scenario
    :param nb_tasks: Number of tasks if we do not want to be set automatically
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    :param filename_hash_indexes: file to save scenarios indexes and reload them after to win some time and computation
    :param split_task: Define if the task split will be automatic by clusterization ("auto") of hashes
    or manually balanced ("balanced"). NB: if split_task == "balanced" nb_tasks can not be None.
    """

    def __init__(
            self,
            cl_dataset: _ContinuumDataset,
            hash_name,
            nb_tasks=None,
            transformations: Union[List[Callable], List[List[Callable]]] = None,
            filename_hash_indexes: Optional[str] = None,
            split_task="balanced"
    ) -> None:
        self.hash_name = hash_name
        self.split_task = split_task
        self._nb_tasks = nb_tasks

        if self.hash_name not in ["AverageHash", "Phash", "PhashSimple", "DhashH", "DhashV", "Whash", "ColorHash"
                                  ]: # , "CropResistantHash"
            AssertionError(f"{self.hash_name} is not a hash_name available.")
        if self.split_task not in ["balanced", "auto"]:
            AssertionError(f"{self.split_task} is not a data_split parameter available.")
        if split_task == "balanced" and nb_tasks is None:
            AssertionError(f"self.data_split is {self.split_task} the nb_tasks should be set.")

        self.data_type = cl_dataset.data_type
        self.filename_hash_indexes = filename_hash_indexes

        # "CropResistantHash" does not work yet
        # if self.hash_name == "CropResistantHash":
        #     # auto (kmeans) does not work with hask format of CropResistantHash
        #     self.split_task = "balanced"

        x, y, t = self.generate_task_ids(cl_dataset)
        cl_dataset = InMemoryDataset(x, y, t, data_type=self.data_type)
        super().__init__(cl_dataset=cl_dataset, transformations=transformations)

    def process_for_hash(self, x):
        """"preprocess data for hashing functions"""
        if self.data_type == "image_array":
            im = Image.fromarray(x.astype("uint8"))
        elif self.data_type == "image_path":
            im = Image.open(x).convert("RGB")
        else:
            raise NotImplementedError(f"data_type -- {self.data_type}"
                                      f" -- Not implemented or not Compatible")

        return im

    def hash_func(self, x):
        ''''Hash one image and return hash'''

        x = self.process_for_hash(x)

        if self.hash_name == "AverageHash":
            hash_value = imagehash.average_hash(x, hash_size=8, mean=np.mean)
        elif self.hash_name == "Phash":
            hash_value = imagehash.phash(x, hash_size=8, highfreq_factor=4)
        elif self.hash_name == "PhashSimple":
            hash_value = imagehash.phash_simple(x, hash_size=8, highfreq_factor=4)
        elif self.hash_name == "DhashH":
            hash_value = imagehash.dhash(x)
        elif self.hash_name == "DhashV":
            hash_value = imagehash.dhash_vertical(x)
        elif self.hash_name == "Whash":
            hash_value = imagehash.whash(x,
                                         hash_size=8,
                                         image_scale=None,
                                         mode='haar',
                                         remove_max_haar_ll=True)
        elif self.hash_name == "ColorHash":
            hash_value = imagehash.colorhash(x, binbits=3)
        elif self.hash_name == "CropResistantHash": # does not work yet
            hash_value = imagehash.crop_resistant_hash(x,
                                                       hash_func=None,
                                                       limit_segments=None,
                                                       segment_threshold=128,
                                                       min_segment_size=500,
                                                       segmentation_image_size=300
                                                       )
        else:
            raise NotImplementedError(f"Hash Name -- {self.hash_name} -- Unknown")

        return str(hash_value)

    def get_task_ids(self, x):
        '''Return the task id vectors corresponding to the parameters settings '''

        if self.split_task == "balanced":
            # In this case: create task ids with a fix set of tasks with a balanced amount of data
            assert self._nb_tasks is not None
            nb_examples = len(x)
            task_ids = np.ones(nb_examples) * (self._nb_tasks - 1)
            example_per_tasks = nb_examples // self._nb_tasks
            perfect_balance_task_ids = np.arange(self._nb_tasks).repeat(example_per_tasks)
            task_ids[:len(perfect_balance_task_ids)] = perfect_balance_task_ids

            # examples from len(perfect_balance_task_ids) to len(task_ids) are put into last tasks
        elif self._nb_tasks is not None:
            # In this case: create task ids with a fix set of tasks with a amount of data automatically set

            # we use KMeans from scikit learn to make hash coherent tasks with a fixed number of task

            # reduce data size for clustering
            pca = PCA(n_components=2)
            reduc_data = pca.fit_transform(x)

            # we use kmeans from scikit learn to create coherent clusters
            task_ids = KMeans(n_clusters=self._nb_tasks).fit_predict(reduc_data)
        else:
            # In this case: create task ids with an automatically set  number of tasks
            # with a amount of data automatically set.

            # we use MeanShift from scikit learn to automatically set the number of task
            # and make hash coherent tasks

            # reduce data size for clustering
            pca = PCA(n_components=2)
            reduc_data = pca.fit_transform(x)

            bandwidth = 5
            task_ids = None
            while (bandwidth > 0.5):
                task_ids = MeanShift(bandwidth=bandwidth, bin_seeding=True).fit_predict(reduc_data)
                if len(np.unique(task_ids)) > 3:
                    # we would like more that 3 tasks if possible
                    break
                else:
                    # reduce the bandwidth if there is not enough tasks
                    bandwidth = bandwidth * 0.75

            self._nb_tasks = len(np.unique(task_ids))
            if not self._nb_tasks > 1:
                AssertionError("The number of task is expected to be more than one.")

        assert len(np.unique(task_ids)) > 1, print(np.unique(task_ids))

        return task_ids

    def get_list_hash_ids(self, x):
        '''Compute hash for all data points in x and return a list of all hash in the same order '''

        # multithread hash evaluation without changing list order
        with Pool(min(8, cpu_count())) as p:
            list_hash = p.map(self.hash_func, list(x))
        return list_hash

    def generate_task_ids(self, cl_dataset):
        ''''This function handle the generation of task id, either by reloading self.filename_hash_indexes
        or by regenerating a task ids vector with self.get_task_ids(...)'''
        x, y, _ = cl_dataset.get_data()

        if self.filename_hash_indexes is not None and os.path.exists(self.filename_hash_indexes):
            print(f"Loading previously saved sorted indexes ({self.filename_hash_indexes}).")
            tuple_indexes_hash = np.load(self.filename_hash_indexes, allow_pickle=True)
            task_ids = tuple_indexes_hash[0].astype(int)
            sort_indexes = tuple_indexes_hash[1].astype(int)
            vectorized_list_hash = tuple_indexes_hash[2]

            assert len(sort_indexes) == len(vectorized_list_hash), print(
                f"sort_indexes {len(sort_indexes)} - list_hash {len(vectorized_list_hash)}")

            x = x[sort_indexes]
            y = y[sort_indexes]
        else:
            list_hash = self.get_list_hash_ids(x)

            vectorized_list_hash = get_array_list(list_hash)

            # arbitrary sorting based on vectorized hash norm
            sort_indexes = sort_hash(vectorized_list_hash)

            x = x[sort_indexes]
            y = y[sort_indexes]
            ordered_hash = np.array(vectorized_list_hash)[sort_indexes]
            task_ids = self.get_task_ids(ordered_hash)

            # save eventually sort_indexes for later use and gain of time
            if self.filename_hash_indexes is not None:
                np.save(self.filename_hash_indexes, [task_ids, sort_indexes, vectorized_list_hash], allow_pickle=True)

        if not len(task_ids) == len(y):
            AssertionError(f"task_ids {len(task_ids)} - y {len(y)} should be equal")

        return x, y, task_ids

    # nothing to do in the setup function
    def _setup(self, nb_tasks: int) -> int:
        return nb_tasks



import warnings
from typing import Callable, List, Optional, Union

import numpy as np

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import _BaseScenario


class InstanceIncremental(_BaseScenario):
    """Continual Loader, generating instance incremental consecutive tasks.

    Scenario: Classes are always the same but instances change (NI scenario)

    :param cl_dataset: A continual dataset.
    :param nb_tasks: The scenario number of tasks.
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    :param random_seed: A random seed to init random processes.
    """

    def __init__(
        self,
        cl_dataset: _ContinuumDataset,
        nb_tasks: Optional[int] = None,
        transformations: Union[List[Callable], List[List[Callable]]] = None,
        random_seed: int = 1
    ):
        super().__init__(cl_dataset=cl_dataset, nb_tasks=nb_tasks, transformations=transformations)

        self._random_state = np.random.RandomState(seed=random_seed)

        self._nb_tasks = self._setup(nb_tasks)

    def _setup(self, nb_tasks: Optional[int]) -> int:
        x, y, t = self.cl_dataset.get_data()

        if nb_tasks is not None and nb_tasks > 0:  # If the user wants a particular nb of tasks
            task_ids = _split_dataset(y, nb_tasks)
            self.dataset = (x, y, task_ids)
        elif t is not None:  # Otherwise use the default task ids if provided by the dataset
            self.dataset = (x, y, t)
            nb_tasks = len(np.unique(t))
        else:
            raise Exception(f"The dataset ({self.cl_dataset}) doesn't provide task ids, "
                            f"you must then specify a number of tasks, not ({nb_tasks}.")

        return nb_tasks


def _split_dataset(y, nb_tasks):
    nb_per_class = np.bincount(y)
    nb_per_class_per_task = nb_per_class / nb_tasks

    if (nb_per_class_per_task < 1.).all():
        raise Exception(f"Too many tasks ({nb_tasks}) for the amount of data "
                        "leading to empty tasks.")
    if (nb_per_class_per_task <= 1.).any():
        warnings.warn(
            f"Number of tasks ({nb_tasks}) is too big resulting in some tasks"
            " without all classes present."
        )

    n = nb_per_class_per_task.astype(np.int64)
    t = np.zeros((len(y),))

    for class_id, nb in enumerate(n):
        t_class = np.zeros((nb_per_class[class_id],))
        for task_id in range(nb_tasks):
            t_class[task_id * nb:(task_id + 1) * nb] = task_id

        t[y == class_id] = t_class

    return t



import os
from typing import Tuple

import pandas as pd
import numpy as np

from continuum.datasets.base import _ContinuumDataset
from continuum.download import download_file_from_google_drive, untar


class CUB200(_ContinuumDataset):
    # initial code taken from https://github.com/TDeVries/cub2011_dataset
    base_folder = "CUB_200_2011/images"

    def __init__(self, data_path, train: bool = True, download: bool = True):
        data_path = os.path.expanduser(data_path)
        self._attributes = None
        super().__init__(data_path, train, download)

    @property
    def attributes(self):
        if self._attributes is None:
            att = np.loadtxt(
                os.path.join(
                    self.data_path, "CUB_200_2011",
                    "attributes", "class_attribute_labels_continuous.txt"
                )
            )
            self._attributes = att / np.linalg.norm(att, axis=-1, keepdims=True)

        return self._attributes

    @property
    def data_type(self):
        return "image_path"

    def _download(self):
        if not os.path.exists(os.path.join(self.data_path, "CUB_200_2011")):
            tgz_path = os.path.join(self.data_path, "CUB_200_2011.tgz")

            if not os.path.exists(tgz_path):
                print("Downloading tgz archive...", end=' ')
                download_file_from_google_drive(
                    "1hbzc_P1FuxMkcabkgn9ZKinBwW683j45",
                    tgz_path
                )
                print('Done!')

            print('Extracting archive...', end=' ')
            untar(tgz_path)
            print('Done!')

    def _load_metadata(self):
        images = pd.read_csv(
            os.path.join(self.data_path, "CUB_200_2011", "images.txt"),
            sep=" ",
            names=["img_id", "filedata_path"])

        image_class_labels = pd.read_csv(
            os.path.join(self.data_path, "CUB_200_2011", "image_class_labels.txt"),
            sep=" ",
            names=["img_id", "target"])
        train_test_split = pd.read_csv(
            os.path.join(self.data_path, "CUB_200_2011", "train_test_split.txt"),
            sep=" ",
            names=["img_id", "is_training_img"])

        data = images.merge(image_class_labels, on="img_id")
        self.data = data.merge(train_test_split, on="img_id")

        if self.train:
            self.data = self.data[self.data.is_training_img == 1]
        else:
            self.data = self.data[self.data.is_training_img == 0]

        x = os.path.join(self.data_path, "CUB_200_2011", "images") + "/" + np.array(data["filedata_path"])
        y = np.array(data["target"]) - 1  # Targets start at 1 by default, so shift to 0

        self.dataset = [x, y, None]

    def _check_integrity(self):
        try:
            self._load_metadata()
        except Exception:
            return False

        for index, row in self.data.iterrows():
            filedata_path = os.path.join(self.data_path, self.base_folder, row.filedata_path)
            if not os.path.isfile(filedata_path):
                print(filedata_path)
                return False
        return True

    def __len__(self):
        """Len measures the number of data point from csv data."""
        return len(self.data)

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. Remove existing file and download again, "
                "or try to download the dataset manually "
                "at http://www.vision.caltech.edu/visipedia/CUB-200-2011.html")

        return self.dataset



import numpy as np
import pytest
from torch.utils.data import DataLoader


from continuum.scenarios import ClassIncremental
from continuum.datasets import MNIST, CIFAR10, CIFAR100, KMNIST, FashionMNIST, InMemoryDataset
from torchvision.transforms import transforms


# yapf: disable


@pytest.mark.slow
@pytest.mark.parametrize("dataset, increment", [(MNIST, 5),
                                                (KMNIST, 2),
                                                (FashionMNIST, 1),
                                                (CIFAR10, 2),
                                                (CIFAR100, 10)])
def test_with_dataset_simple_increment(tmpdir, dataset, increment):
    dataset = dataset(data_path=tmpdir, download=True, train=True)
    scenario = ClassIncremental(cl_dataset=dataset,
                                increment=increment,
                                transformations=[transforms.ToTensor()]
                                )

    for task_id, taskset in enumerate(scenario):
        classes = taskset.get_classes()

        assert len(classes) == increment

        # check if there is continuity in classes by default
        assert len(classes) == (classes.max() - classes.min() + 1)


@pytest.mark.slow
@pytest.mark.parametrize("dataset, increment", [(MNIST, [5, 1, 1, 3]),
                                                (KMNIST, [2, 2, 4, 2]),
                                                (FashionMNIST, [1, 2, 1, 2, 1, 2, 1]),
                                                (CIFAR10, [2, 2, 2, 2, 2]),
                                                (CIFAR100, [50, 10, 20, 20])])
def test_with_dataset_composed_increment(tmpdir, dataset, increment):
    dataset = dataset(data_path=tmpdir, download=True, train=True)
    scenario = ClassIncremental(cl_dataset=dataset,
                                increment=increment,
                                transformations=[transforms.ToTensor()]
                                )

    for task_id, taskset in enumerate(scenario):
        classes = taskset.get_classes()

        assert len(classes) == increment[task_id]

        # check if there is continuity in classes by default
        assert len(classes) == (classes.max() - classes.min() + 1)


NB_CLASSES=10

@pytest.fixture
def fake_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(NB_CLASSES):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    return InMemoryDataset(x_train, y_train)


@pytest.mark.parametrize("class_order", [
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    [3, 9, 2, 0, 4, 5, 9, 7, 6, 1],
])
def test_taskid(fake_data, class_order):
    scenario = ClassIncremental(
        cl_dataset=fake_data,
        increment=2
    )

    for task_id, taskset in enumerate(scenario):
        loader = DataLoader(taskset, batch_size=32)

        for x, y, t in loader:
            assert t[0].item() == task_id
            assert (t == task_id).all()

def test_nb_classes(fake_data):
    scenario = ClassIncremental(
        cl_dataset=fake_data,
        increment=2
    )

    assert scenario.nb_classes == NB_CLASSES
    assert (scenario.classes == np.arange(NB_CLASSES)).all()


import os
import json
from typing import Tuple, Optional

from torchvision import datasets as torchdata
import numpy as np

from continuum.datasets import _ContinuumDataset
from continuum.download import download, unzip


class Stream51(_ContinuumDataset):
    """Stream51 dataset.

    Reference:
        * Stream-51: Streaming Classification and Novelty Detection From Videos
          Roady, Ryne and Hayes, Tyler L. and Vaidya, Hitesh and Kanan, Christopher
          CVPR Workshops 2020

    Official implementation here: https://github.com/tyler-hayes/Stream-51

    :param data_path: The folder path containing the data.
    :param train: Train or Test mode.
    :param download: Auto-download of the dataset if it doesn't exist.
    :param crop: Uses the bounding boxes to crop objects intead of taking whole images.
    :param ratio: A ratio factor to take a bit more pixels than just the bounding box.
    :param task_criterion: Criterion to split instances into tasks (other than class).
                           Can be either "clip" or "video". This parameter only
                           has effect in the InstanceIncremental scenario. If you
                           use the ClassIncremental scenario, only the class ids
                           will be taken in account to create tasks.
    """

    url = "http://klab.cis.rit.edu/files/Stream-51.zip"

    def __init__(
            self,
            data_path: str = "",
            train: bool = True,
            download: bool = True,
            crop: bool = True,
            ratio: float = 1.1,
            task_criterion: str = "clip"
    ):
        if task_criterion not in ("clip", "video"):
            raise ValueError(
                f"Invalid task criterion: {task_criterion}, expect 'clip' or 'video'."
            )

        super().__init__(data_path=data_path, train=train, download=download)

        self.crop = crop
        self.ratio = ratio
        self.task_criterion = task_criterion
        self._bounding_boxes = None

    def _download(self):
        path = os.path.join(self.data_path, "Stream-51")
        if not os.path.exists(f"{path}.zip"):
            download(self.url, self.data_path)
        if not os.path.exists(path):
            unzip(f"{path}.zip")

        print("Stream51 downloaded.")

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x, y, t, b = self._parse_json()

        if self.crop:
            self._set_bounding_boxes(b)

        return x, y, t

    def _set_bounding_boxes(self, bounding_boxes: np.ndarray):
        """Format the bounding boxes in order to enlarge them through ratio param.

        See official code about it, there:
        https://github.com/tyler-hayes/Stream-51/blob/master/experiments/StreamDataset.py#L139

        :param bounding_boxes: A list of bounding boxes as provided by the dataset.
        """
        formatted_bounding_boxes = []

        for bbox in bounding_boxes:
            cw = bbox[0] - bbox[1]
            ch = bbox[2] - bbox[3]
            center = [int(bbox[1] + cw / 2), int(bbox[3] + ch / 2)]
            formatted_bounding_boxes.append([
                int(center[0] - (cw * self.ratio / 2)),  # x1
                int(center[1] - (ch * self.ratio / 2)),  # y1
                int(center[0] + (cw * self.ratio / 2)),  # x2
                int(center[1] + (ch * self.ratio / 2)),  # y2
            ])

        self._bounding_boxes = np.array(formatted_bounding_boxes)

    @property
    def bounding_boxes(self):
        return self._bounding_boxes

    @property
    def data_type(self) -> str:
        return "image_path"

    def _parse_json(self):
        if self.train:
            path = os.path.join(self.data_path, "Stream-51", "Stream-51_meta_train.json")
        else:
            path = os.path.join(self.data_path, "Stream-51", "Stream-51_meta_test.json")

        # X: path, Y: class id, T: task id (aka clip or video id), B: bounding box
        x, y, t, b = [], [], [], []
        with open(path) as f:
            data = json.load(f)

        for line in data:
            # line : [class_id, clip_num, video_num, frame_num, img_shape, bbox, file_loc]
            # clip_num, video_num and frame_num relative
            x.append(os.path.join(self.data_path, "Stream-51", line[-1]))
            y.append(line[0])
            if self.train:
                if self.task_criterion == "clip":
                    num = line[1] #clip_num
                else:
                    num = line[2] #video_num
                t.append(num)
            else:
                t.append(0)
            b.append(line[-2])

        t = np.array(t)
        for index, unique_task_id in enumerate(np.unique(t)):
            t[t == unique_task_id] = index

        return np.array(x), np.array(y), t, np.array(b)



import numpy as np
import pytest
import os

from continuum.datasets import MNIST, CIFAR10, CIFAR100, KMNIST, FashionMNIST, TinyImageNet200
from continuum.scenarios import Rotations
from continuum.scenarios import Permutations
from continuum.scenarios import ClassIncremental
from continuum.datasets import MNISTFellowship


@pytest.mark.slow
@pytest.mark.parametrize("dataset, name, shape", [(MNIST, "MNIST", [28, 28, 1]),
                                                  (KMNIST, "KMNIST", [28, 28, 1]),
                                                  (FashionMNIST, "FashionMNIST", [28, 28, 1]),
                                                  (CIFAR10, "CIFAR10", [32, 32, 3]),
                                                  (CIFAR100, "CIFAR100", [32, 32, 3]),
                                                  (TinyImageNet200, "TinyImageNet200", [64, 64, 3])])
def test_visualization_ClassIncremental(tmpdir, dataset, name, shape):
    increment = 2
    if name == "CIFAR100":
        increment = 20
    if name == "TinyImageNet200":
        increment = 40
    scenario = ClassIncremental(cl_dataset=dataset(data_path=tmpdir, download=True, train=True),
                                increment=increment)

    folder = "tests/samples/class_incremental/"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for task_id, taskset in enumerate(scenario):
        taskset.plot(path=folder,
                     title="{}_ClassIncremental_{}.jpg".format(name, task_id),
                     nb_samples=100,
                     shape=shape)


'''
Test the visualization with three tasks for rotations tasks
'''
@pytest.mark.slow
def test_visualization_rotations(tmpdir):
    scenario = Rotations(cl_dataset=MNIST(data_path=tmpdir, download=True, train=True),
                         nb_tasks=3,
                         list_degrees=[0, 45, 92])

    folder = "tests/samples/rotations/"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for task_id, taskset in enumerate(scenario):
        taskset.plot(path=folder,
                     title="MNIST_Rotation_{}.jpg".format(task_id),
                     nb_samples=100,
                     shape=[28, 28, 1])


'''
Test the visualization with three tasks for permutations tasks
'''
@pytest.mark.slow
def test_visualization_permutations(tmpdir):
    scenario = Permutations(cl_dataset=MNIST(data_path=tmpdir, download=True, train=True),
                            nb_tasks=3,
                            seed=0)

    folder = "tests/samples/permutations/"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for task_id, taskset in enumerate(scenario):
        taskset.plot(path=folder,
                     title="MNIST_Permutations_{}.jpg".format(task_id),
                     nb_samples=100,
                     shape=[28, 28, 1])


'''
Test the visualization with three tasks for incremental tasks
'''
@pytest.mark.slow
def test_visualization_MNISTFellowship(tmpdir):
    cl_dataset = MNISTFellowship(data_path=tmpdir, download=True, train=True)
    scenario = ClassIncremental(cl_dataset=cl_dataset,
                                increment=10)

    folder = "tests/samples/fellowship/"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for task_id, taskset in enumerate(scenario):
        taskset.plot(path=folder,
                     title="MNISTFellowship_Incremental_{}.jpg".format(task_id),
                     nb_samples=100,
                     shape=[28, 28, 1])



import os

from PIL import Image
import matplotlib.pyplot as plt
import numpy as np


def plot_samples(dataset, title="", path=None, nb_samples=100, shape=None, data_type="image_array"):
    batch, y, _ = dataset.get_random_samples(nb_samples)

    if len(y.shape) == 1:
        y, order = y.sort()
        batch = batch[order]

    if path is not None:
        filename = os.path.join(path, title)
    else:
        filename = None

    if shape is None:
        shape = batch[0].shape

    if data_type == "segmentation":
        visualize_segmentation_batch(batch, y, nb_samples, shape, filename)
    else:
        visualize_batch(batch, nb_samples, shape, filename)


def visualize_batch(batch, number, shape, path):
    batch = batch.cpu().data

    image_frame_dim = int(np.floor(np.sqrt(number)))

    if shape[2] == 1:
        data_np = batch.numpy().reshape(number, shape[0], shape[1], shape[2])
        save_images(
            data_np[:image_frame_dim * image_frame_dim, :, :, :],
            [image_frame_dim, image_frame_dim], path
        )
    elif shape[2] == 3:
        data = batch.numpy().reshape(number, shape[2], shape[1], shape[0])
        data = img_stretch(data)
        make_samples_batch(data[:number], number, path)
    else:
        save_images(
            batch[:image_frame_dim * image_frame_dim, :, :, :], [image_frame_dim, image_frame_dim],
            path
        )


def visualize_segmentation_batch(images, segmaps, number, shape, path):
    images, segmaps = images.numpy(), segmaps.numpy()
    images = img_stretch(images)
    make_samples_segmentation_batch(images[:number], segmaps[:number], number, path)


def save_images(images, size, image_path):
    images = np.array(images)
    if images.shape[1] in (1, 3):  # When channel axis is before spatial axis.
        images = images.transpose(0, 2, 3, 1)

    image = np.squeeze(merge(images, size))
    image -= np.min(image)
    image /= np.max(image) + 1e-12
    image = 255 * image  # Now scale by 255
    image = image.astype(np.uint8)

    if image_path is not None:
        return Image.fromarray(image).save(image_path)
    else:
        cmap = None
        if len(image.shape) == 2:
            cmap = "gray"
        plt.axis('off')
        plt.imshow(image, cmap=cmap)


def merge(images, size):
    img = None
    h, w = images.shape[1], images.shape[2]
    if (images.shape[3] in (3, 4)):
        c = images.shape[3]
        img = np.zeros((h * size[0], w * size[1], c))
        for idx, image in enumerate(images):
            i = idx % size[1]
            j = idx // size[1]
            img[j * h:j * h + h, i * w:i * w + w, :] = image
    elif images.shape[3] == 1:
        img = np.zeros((h * size[0], w * size[1]))
        for idx, image in enumerate(images):
            i = idx % size[1]
            j = idx // size[1]
            img[j * h:j * h + h, i * w:i * w + w] = image[:, :, 0]
    else:
        raise ValueError(
            'in merge(images,size) images parameter '
            'must have dimensions: BxHxW or BxHxWx3 or BxHxWx4 '
            f'not {images.shape}'
        )

    return img


def img_stretch(img):
    img = img.astype(float)
    img -= np.min(img)
    img /= np.max(img) + 1e-12
    return img


def make_samples_batch(images, batch_size, path):
    plt.figure()
    images, batch_size = _make_square_group(images, batch_size)
    fig, ax = plt.subplots(figsize=(batch_size, batch_size))
    ax.axis('off')
    ax.imshow(images, interpolation='nearest')
    ax.grid()
    ax.set_xticks([])
    ax.set_yticks([])

    if path is not None:
        fig.savefig(path, bbox_inches='tight', pad_inches=0)
    else:
        plt.show()

    plt.close(fig)
    plt.close()


def make_samples_segmentation_batch(images, labels, batch_size, path):
    images, _ = _make_square_group(images, batch_size)
    labels, _ = _make_square_group(labels, batch_size)

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))

    axes[0].axis('off')
    axes[0].imshow(images, interpolation='nearest')
    axes[0].grid()
    axes[0].set_xticks([])
    axes[0].set_yticks([])

    # TODO: Add color map for other dataset than VOC
    nclasses = 21
    row_size = images.shape[-1]
    col_size = images.shape[-2]
    cmap = color_map()[:, np.newaxis, :]
    labels = labels[:, :, np.newaxis]
    new_im = np.dot(labels == 0, cmap[0])
    for i in range(1, cmap.shape[0]):
        new_im += np.dot(labels == i, cmap[i])
    labels = new_im

    axes[1].axis('off')
    axes[1].imshow(labels, interpolation='nearest')
    axes[1].grid()
    axes[1].set_xticks([])
    axes[1].set_yticks([])

    if path is not None:
        fig.savefig(path, bbox_inches='tight', pad_inches=0)
    else:
        plt.show()

    plt.close(fig)
    plt.close()


def color_map(N=256, normalized=False):
    def bitget(byteval, idx):
        return ((byteval & (1 << idx)) != 0)

    dtype = 'float32' if normalized else 'uint8'
    cmap = np.zeros((N, 3), dtype=dtype)
    for i in range(N):
        r = g = b = 0
        c = i
        for j in range(8):
            r = r | (bitget(c, 0) << 7-j)
            g = g | (bitget(c, 1) << 7-j)
            b = b | (bitget(c, 2) << 7-j)
            c = c >> 3

        cmap[i] = np.array([r, g, b])

    cmap = cmap/255 if normalized else cmap
    return cmap



def _make_square_group(images, batch_size):
    nb_dim = len(images.shape)
    if nb_dim == 3:
        images = np.repeat(images[:, None], 3, axis=1)

    batch_size_sqrt = int(np.sqrt(batch_size))
    input_channel = images[0].shape[0]
    input_dim = images[0].shape[1]

    if nb_dim != 3:
        images = np.clip(images, 0, 1)
    images = images[:batch_size_sqrt ** 2]

    images = np.rollaxis(
        images.reshape((batch_size_sqrt, batch_size_sqrt, input_channel, input_dim, input_dim)),
        2, 5
    )
    images = images.swapaxes(2, 1)
    images = images.reshape((batch_size_sqrt * input_dim, batch_size_sqrt * input_dim, input_channel))

    if nb_dim == 3:
        images = images[..., 0]

    return images, batch_size_sqrt





import numpy as np
import pytest

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental, InstanceIncremental


def gen_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


# yapf: disable

@pytest.mark.parametrize("index,classes", [
    (0, [0, 1]),
    (1, [2, 3]),
    (slice(0, 2, 1), [0, 1, 2, 3]),
    (slice(0, 2), [0, 1, 2, 3]),
    (slice(2), [0, 1, 2, 3]),
    (slice(1, 3), [2, 3, 4, 5]),
    (slice(0, 10, 2), [0, 1, 4, 5, 8, 9]),
    (-1, [8, 9]),
    (-5, [0, 1]),
    (slice(-1, -3, -1), [6, 7, 8, 9]),
    (-6, [8, 9]),  # full loop
    (-7, [6, 7]),  # full loop
    (-20, [0, 1]),  # full loop
])
def test_slicing_nc(index, classes):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=2)
    taskset = scenario[index]
    targets = np.sort(np.unique(taskset._y))
    assert len(targets) == len(classes)
    assert (targets == np.array(classes)).all(), (targets, classes)


@pytest.mark.parametrize("start_index,classes", [
    (0, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
    (3, [6, 7, 8, 9]),
    (4, [8, 9]),
])
def test_slicing_nc_no_end(start_index, classes):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=2)
    taskset = scenario[start_index:]
    targets = np.sort(np.unique(taskset._y))
    assert len(targets) == len(classes)
    assert (targets == np.array(classes)).all(), (targets, classes)


@pytest.mark.parametrize("start,end", [
    (12, 100),
    (0, 0),
    (3, 3)
])
def test_slicing_empty(start, end):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=2)

    has_failed = False
    try:
        taskset = scenario[start_index:]
    except:
        has_failed = True

    assert has_failed


def test_slicing_nc_no_index():
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=2)
    taskset = scenario[:]
    targets = np.sort(np.unique(taskset._y))

    classes = list(range(10))

    assert len(targets) == len(classes)
    assert (targets == np.array(classes)).all(), (targets, classes)


@pytest.mark.parametrize("list_tasks", [
    np.arange(10),
    np.arange(5, 10),
    np.arange(3, 10, 2),
    np.arange(9, 0, -2),
    np.arange(0, 10, 2),
    list(np.arange(0, 10, 2)),
    list(np.arange(5, 10))
])
def test_slicing_list(list_tasks):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)
    taskset = scenario[list_tasks]
    targets = np.sort(np.unique(taskset._y))
    assert len(targets) == len(list_tasks), print(f"{len(targets)} - vs - {len(list_tasks)}")



import json
import os
from typing import Dict, List, Tuple

import numpy as np

from continuum import download
from continuum.datasets.base import _ContinuumDataset


class MultiNLI(_ContinuumDataset):
    """Continuum version of the MultiNLI dataset.

    References:
        * A Broad-Coverage Challenge Corpus for Sentence Understanding through Inference
          Williams, Nangia, and Bowman
          ACL 2018
        * Progressive Memory Banks for Incremental Domain Adaptation
          Asghar & Mou
          ICLR 2020

    The dataset is based on the NLI task.
    For each example, two sentences are given. The goal is to determine whether
    this pair of sentences has:
    - Opposite meaning (contradiction)
    - Similar meaning (entailment)
    - no relation to each other (neutral)

    :param data_path: The folder extracted from the official zip file.
    :param download: An option useless in this case.
    """

    data_url = "https://cims.nyu.edu/~sbowman/multinli/multinli_1.0.zip"

    def __init__(self, data_path: str = "", train: bool = True, download: bool = True) -> None:
        super().__init__(data_path=data_path, train=train, download=download)

    def _download(self):
        if os.path.exists(os.path.join(self.data_path, "multinli_1.0")):
            print("Dataset already extracted.")
        else:
            path = download.download(self.data_url, self.data_path)
            download.unzip(path)
            print("Dataset extracted.")

    @property
    def data_type(self) -> str:
        return "text"

    @property
    def transformations(self):
        return []

    def original_targets(self) -> List[str]:
        return ["contradiction", "entailment", "neutral"]

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Generate the MultiNLI data.

        The dataset has several domains, but always the same targets
        ("contradiction", "entailment", "neutral"). 5 domains are allocated for
        the train set ("fiction", "government", "slate", "telephone", "travel"),
        and 5 to the test set ("facetoface", "letters", "nineeleven", "oup",
        "verbatim").

        While the train is given different task id for each domain, the test set
        always has a dummy 0 domain id, as it is supposed to be fixed.
        """
        texts, targets, genres = [], [], []

        available_targets = ["contradiction", "entailment", "neutral"]
        available_genres = [
            "fiction",
            "government",
            "slate",
            "telephone",
            "travel",  # /train
            "facetoface",
            "letters",
            "nineeleven",
            "oup",
            "verbatim"  # /test
        ]

        if self.train:
            json_path = os.path.join(self.data_path, "multinli_1.0", "multinli_1.0_train.jsonl")
        else:
            json_path = os.path.join(
                self.data_path, "multinli_1.0", "multinli_1.0_dev_mismatched.jsonl"
            )

        with open(json_path) as f:
            for line in f:
                line_parsed: Dict[str, str] = json.loads(line)

                if line_parsed["gold_label"] not in available_targets:
                    continue  # A few cases exist w/o targets.

                texts.append((line_parsed["sentence1"], line_parsed["sentence2"]))
                targets.append(available_targets.index(line_parsed["gold_label"]))

                if self.train:  # We add a new domain id for the train set.
                    genres.append(available_genres.index(line_parsed["genre"]))
                else:  # Test set is fixed, therefore we artificially give a unique domain.
                    genres.append(0)

        texts = np.array(texts)
        targets = np.array(targets)
        genres = np.array(genres)

        return texts, targets, genres



import pytest

from continuum import datasets as cont_datasets
from torchvision.datasets import EMNIST, KMNIST
from continuum.datasets import PyTorchDataset

ATTRS = ["get_data", "_download"]


@pytest.mark.parametrize("dataset_name", [d for d in dir(cont_datasets) if d[0].isupper()])
def test_has_attr(dataset_name):
    d = getattr(cont_datasets, dataset_name)

    for attr in ATTRS:
        assert hasattr(d, attr), (dataset_name, attr)


@pytest.mark.slow
def test_PytorchDataset_EMNIST(tmpdir):
    dataset_train = PyTorchDataset(tmpdir, dataset_type=EMNIST, train=True, download=True, split='letters')


@pytest.mark.slow
def test_PytorchDataset_KMNIST(tmpdir):
    dataset_train = PyTorchDataset(tmpdir, dataset_type=KMNIST, train=True, download=True)


import abc
import os
from typing import List, Tuple, Union
import warnings

import numpy as np
from torchvision import datasets as torchdata
from torchvision import transforms

from continuum.transforms.segmentation import ToTensor as ToTensorSegmentation


class _ContinuumDataset(abc.ABC):

    def __init__(self, data_path: str = "", train: bool = True, download: bool = True) -> None:
        self.data_path = os.path.expanduser(data_path)
        self.download = download
        self.train = train

        if not os.path.exists(self.data_path):
            os.makedirs(self.data_path)

        if self.download:
            self._download()

        if self.data_type not in ("image_array", "image_path", "text", "tensor", "segmentation"):
            raise NotImplementedError(f"Dataset's data_type ({self.data_type}) is not supported.")

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        pass

    def _download(self):
        pass

    @property
    def class_order(self) -> Union[None, List[int]]:
        return None

    @property
    def need_class_remapping(self) -> bool:
        """Flag for method `class_remapping`."""
        return False

    def class_remapping(self, class_ids: np.ndarray) -> np.ndarray:
        """Optional class remapping.

        Used for example in PermutedMNIST, cf transformed.py;

        :param class_ids: Original class_ids.
        :return: A remapping of the class ids.
        """
        return class_ids

    @property
    def data_type(self) -> str:
        return "image_array"

    @property
    def transformations(self):
        """Default transformations if nothing is provided to the scenario."""
        if self.data_type == "segmentation":
            return [ToTensorSegmentation()]
        return [transforms.ToTensor()]

    @property
    def bounding_boxes(self) -> List:
        """Returns a bounding box (x1, y1, x2, y2) per sample if they need to be cropped."""
        return None

    @property
    def attributes(self) -> np.ndarray:
        """Returns normalized attributes for all class if available.

        Those attributes can often be found in dataset used for Zeroshot such as
        CUB200, or AwA. The matrix shape is (nb_classes, nb_attributes), and it
        has been L2 normalized along side its attributes dimension.
        """
        return None


class _SemanticSegmentationDataset(_ContinuumDataset):
    """Base class for segmentation-based dataset."""

    @property
    def data_type(self) -> str:
        return "segmentation"


class PyTorchDataset(_ContinuumDataset):
    """Continuum version of torchvision datasets.
    :param dataset_type: A Torchvision dataset, like MNIST or CIFAR100.
    :param train: train flag
    :param download: download
    """

    # TODO: some datasets have a different structure, like SVHN for ex. Handle it.
    def __init__(
            self, data_path: str = "", dataset_type=None, train: bool = True, download: bool = True, **kwargs):

        if "transform" in kwargs:
            raise ValueError(
                "Don't provide `transform` to the dataset. "
                "You should give those to the scenario."
            )

        super().__init__(data_path=data_path, train=train, download=download)

        self.dataset_type = dataset_type
        self.dataset = self.dataset_type(self.data_path, download=self.download, train=self.train, **kwargs)

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x, y = np.array(self.dataset.data), np.array(self.dataset.targets)

        if 0 not in y:
            # This case can happen when the first class id is 1 and not 0.
            # For example in EMNIST with 'letters' split (WTF right).
            # TODO: We should handle this case in a more generic fashion later.
            warnings.warn("Converting 1-based class ids to 0-based class ids.")
            y -= 1

        return x, y, None


class InMemoryDataset(_ContinuumDataset):
    """Continuum dataset for in-memory data.

    :param x_train: Numpy array of images or paths to images for the train set.
    :param y_train: Targets for the train set.
    :param data_type: Format of the data.
    :param t_train: Optional task ids for the train set.
    """

    def __init__(
            self,
            x: np.ndarray,
            y: np.ndarray,
            t: Union[None, np.ndarray] = None,
            data_type: str = "image_array",
            train: bool = True,
            download: bool = True,
    ):
        super().__init__(train=train, download=download)

        if len(x) != len(y):
            raise ValueError(f"Number of datapoints ({len(x)}) != number of labels ({len(y)})!")
        if t is not None and len(t) != len(x):
            raise ValueError(f"Number of datapoints ({len(x)}) != number of task ids ({len(t)})!")

        self.data = (x, y, t)
        if data_type not in ("image_array", "image_path", "text", "tensor", "segmentation"):
            raise ValueError(f"Unrecognized data_type={data_type} for InMemoryDataset!")
        self._data_type = data_type

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self.data

    @property
    def data_type(self) -> str:
        return self._data_type

    @data_type.setter
    def data_type(self, data_type: str) -> None:
        self._data_type = data_type


class ImageFolderDataset(_ContinuumDataset):
    """Continuum dataset for datasets with tree-like structure.

    :param train_folder: The folder of the train data.
    :param test_folder: The folder of the test data.
    :param download: Dummy parameter.
    """

    def __init__(self, data_path: str, train: bool = True, download: bool = True, data_type: str = "image_path"):
        self.data_path = data_path
        super().__init__(data_path=data_path, train=train, download=download)

        allowed_data_types = ("image_path", "segmentation")
        if data_type not in allowed_data_types:
            raise ValueError(f"Invalid data_type={data_type}, allowed={allowed_data_types}.")
        self._data_type = data_type

    @property
    def data_type(self) -> str:
        return self._data_type

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, Union[None, np.ndarray]]:
        self.dataset = torchdata.ImageFolder(self.data_path)
        return self._format(self.dataset.imgs)

    @staticmethod
    def _format(raw_data: List[Tuple[str, int]]) -> Tuple[np.ndarray, np.ndarray, None]:
        x = np.empty(len(raw_data), dtype="S255")
        y = np.empty(len(raw_data), dtype=np.int16)

        for i, (path, target) in enumerate(raw_data):
            x[i] = path
            y[i] = target

        return x, y, None



import numpy as np
import pytest
import random
import string

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental, ContinualScenario, create_subscenario


def gen_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


def gen_string():
    """"create random string. We plan to use them as it was some path"""
    len_string = 20
    y_train = []
    list_str = []
    for i in range(20):
        value = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(len_string))
        list_str.append([value])
    x_train = np.concatenate(list_str)

    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)
    return x_train, y_train


@pytest.mark.parametrize("list_tasks", [
    np.arange(10),
    np.arange(5, 10),
    np.arange(3, 10, 2),
    np.arange(9, 0, -2),
    np.arange(0, 10, 2),
    list(np.arange(0, 10, 2)),
    list(np.arange(5, 10))
])
def test_slicing_list(list_tasks):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=1)
    subscenario = create_subscenario(scenario, list_tasks)
    assert subscenario.nb_tasks == len(list_tasks), print(f"{len(subscenario)} - vs - {len(list_tasks)}")


@pytest.mark.parametrize("list_tasks", [
    np.arange(10),
    np.arange(5, 10),
    np.arange(3, 10, 2),
    np.arange(9, 0, -2),
])
def test_slicing_list_path_array(list_tasks):
    x_train, y_train = gen_string()
    dummy = InMemoryDataset(x_train, y_train, data_type="image_path")
    scenario = ClassIncremental(dummy, increment=1)
    subscenario = create_subscenario(scenario, list_tasks)
    assert subscenario.nb_tasks == len(list_tasks), print(f"{len(subscenario)} - vs - {len(list_tasks)}")



from typing import List, Tuple

import numpy as np

from continuum.datasets.base import _ContinuumDataset
from continuum.datasets.pytorch import (CIFAR10, KMNIST, MNIST, FashionMNIST)
from continuum.datasets.cifar100 import CIFAR100


class Fellowship(_ContinuumDataset):
    """A dataset concatenating multiple datasets.

    :param datasets: A list of instanciated Continuum dataset.
    :param update_labels: if true, the class id are incremented for each new
                          concatenated datasets, otherwise the class ids are
                          shared, e.g. a Fellowship of CIFAR10+CIFAR100 with
                          update_labels=False will share the first 10 classes.
                          In doubt, let this param to True.
    """
    def __init__(
        self,
        datasets: List[_ContinuumDataset],
        update_labels: bool = True,
    ):
        super().__init__()

        self.update_labels = update_labels
        self.datasets = datasets


    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x, y, t = [], [], []
        class_counter = 0

        for i, dataset in enumerate(self.datasets):
            data = dataset.get_data()

            x.append(data[0])
            if self.update_labels:
                y.append(data[1] + class_counter)
            else:
                y.append(data[1])

            t.append(np.ones(len(data[0])) * i)

            class_counter += len(np.unique(data[1]))

        x = np.concatenate(x)
        y = np.concatenate(y)
        t = np.concatenate(t)

        # There should be by default as much task id as datasets
        assert len(np.unique(t)) == len(self.datasets), f'They should be as much datasets as task ids,' \
                                                        f' we have {len(self.datasets)} datasets vs' \
                                                        f' {len(np.unique(t))} task ids'

        return x, y, t


class MNISTFellowship(Fellowship):

    def __init__(self,
                 data_path: str = "",
                 train: bool = True,
                 download: bool = True,
                 update_labels: bool = True) -> None:
        super().__init__(
            datasets=[
                MNIST(data_path=data_path, train=train, download=download),
                FashionMNIST(data_path=data_path, train=train, download=download),
                KMNIST(data_path=data_path, train=train, download=download)
            ],
            update_labels=update_labels
        )


class CIFARFellowship(Fellowship):

    def __init__(self,
                 data_path: str = "",
                 train: bool = True,
                 download: bool = True,
                 update_labels: bool = True) -> None:
        super().__init__(
            datasets=[
                CIFAR10(data_path=data_path, train=train, download=download),
                CIFAR100(data_path=data_path, train=train, download=download)
            ],
            update_labels=update_labels
        )



import os

import pytest

from continuum.datasets import ImageNet100
from continuum.scenarios import ClassIncremental


nb_images_per_subset = {
    True: 129395,  # train
    False: 5000    # test
}


@pytest.fixture
def ImageNet100Test(tmpdir):
    folder = os.path.join(tmpdir, "imagenet100test")
    os.makedirs(folder)
    return ImageNet100(folder, data_subset=None, download=True, train=False)


@pytest.fixture
def ImageNet100Train(tmpdir):
    folder = os.path.join(tmpdir, "imagenet100train")
    os.makedirs(folder)
    return ImageNet100(folder, data_subset=None, download=True, train=True)


@pytest.mark.parametrize("train", [True, False])
def test_parsing_imagenet100(ImageNet100Train, ImageNet100Test, train):
    dataset = ImageNet100Train if train else ImageNet100Test
    x, y, t = dataset.get_data()

    assert all("train" if train else "test" in path for path in x)



@pytest.mark.parametrize("train", [True, False])
def test_nb_imagenet100(ImageNet100Train, ImageNet100Test, train):
    dataset = ImageNet100Train if train else ImageNet100Test
    x, y, t = dataset.get_data()

    assert len(x) == nb_images_per_subset[train]


@pytest.mark.parametrize("train,div", [
    (True, 1), (True, 2),
    (False, 1), (True, 2)
])
def test_customsubset_imagenet100(ImageNet100Train, ImageNet100Test, train, div):
    dataset = ImageNet100Train if train else ImageNet100Test
    x, y, t = dataset.get_data()

    new_x = x[:len(x) // div]
    new_y = y[:len(y) // div]

    subset = ImageNet100(dataset.data_path, data_subset=(new_x, new_y), download=False, train=train)
    x2, y2, t2 = subset.get_data()

    assert len(x) // div == len(x2)






import abc
from typing import Callable, List, Union

import numpy as np
from torchvision import transforms

from continuum.datasets import _ContinuumDataset
from continuum.tasks import TaskSet
from continuum.transforms.segmentation import Compose as SegmentationCompose


class _BaseScenario(abc.ABC):
    """Abstract loader.

    DO NOT INSTANTIATE THIS CLASS.

    :param cl_dataset: A Continuum dataset.
    :param nb_tasks: The number of tasks to do.
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    """

    def __init__(
            self,
            cl_dataset: _ContinuumDataset,
            nb_tasks: int,
            transformations: Union[List[Callable], List[List[Callable]]] = None
    ) -> None:

        self.cl_dataset = cl_dataset
        self._nb_tasks = nb_tasks
        self.transformations = transformations

        if transformations is None:
            self.transformations = self.cl_dataset.transformations
        if self.cl_dataset.data_type == "segmentation":
            composer = SegmentationCompose
        else:
            composer = transforms.Compose
        if self.transformations is not None and isinstance(self.transformations[0], list):
            # We have list of list of callable, where each sublist is dedicated to
            # a task.
            self.trsf = [composer(trsf) for trsf in self.transformations]
        else:
            self.trsf = composer(self.transformations)

    @abc.abstractmethod
    def _setup(self, nb_tasks: int) -> int:
        raise NotImplementedError

    @property
    def train(self) -> bool:
        """Returns whether we are in training or testing mode.

        This property is dependent on the dataset, not the actual scenario.
        """
        return self.cl_dataset.train

    @property
    def nb_classes(self) -> int:
        """Total number of classes in the whole continual setting."""
        return len(np.unique(self.dataset[1]))  # type: ignore

    @property
    def classes(self) -> int:
        """list of classes in the whole continual setting."""
        return np.unique(self.dataset[1])  # type: ignore

    @property
    def nb_tasks(self) -> int:
        """Number of tasks in the whole continual setting."""
        return len(self)

    def __len__(self) -> int:
        """Returns the number of tasks.

        :return: Number of tasks.
        """
        return self._nb_tasks

    def __iter__(self):
        """Used for iterating through all tasks with the CLLoader in a for loop."""
        self._counter = 0
        return self

    def __next__(self) -> TaskSet:
        """An iteration/task in the for loop."""
        if self._counter >= len(self):
            raise StopIteration
        task = self[self._counter]
        self._counter += 1
        return task

    def __getitem__(self, task_index: Union[int, slice]):
        """Returns a task by its unique index.

        :param task_index: The unique index of a task. As for List, you can use
                           indexing between [0, len], negative indexing, or
                           even slices.
        :return: A train PyTorch's Datasets.
        """
        if isinstance(task_index, slice) and isinstance(self.trsf, list):
            raise ValueError(
                f"You cannot select multiple task ({task_index}) when you have a "
                "different set of transformations per task"
            )

        x, y, t, _ = self._select_data_by_task(task_index)

        return TaskSet(
            x, y, t,
            trsf=self.trsf[task_index] if isinstance(self.trsf, list) else self.trsf,
            data_type=self.cl_dataset.data_type,
            bounding_boxes=self.cl_dataset.bounding_boxes
        )

    def _select_data_by_task(
            self,
            task_index: Union[int, slice, np.ndarray]
    ) -> Union[np.ndarray, np.ndarray, np.ndarray, Union[int, List[int]]]:
        """Selects a subset of the whole data for a given task.

        This class returns the "task_index" in addition of the x, y, t data.
        This task index is either an integer or a list of integer when the user
        used a slice. We need this variable when in segmentation to disangle
        samples with multiple task ids.

        :param task_index: The unique index of a task. As for List, you can use
                           indexing between [0, len], negative indexing, or
                           even slices.
        :return: A tuple of numpy array being resp. (1) the data, (2) the targets,
                 (3) task ids, and (4) the actual task required by the user.
        """
        x, y, t = self.dataset  # type: ignore

        if isinstance(task_index, slice):
            start = task_index.start if task_index.start is not None else 0
            stop = task_index.stop if task_index.stop is not None else len(self) + 1
            step = task_index.step if task_index.step is not None else 1
            task_index = list(range(start, stop, step))
            if len(task_index) == 0:
                raise ValueError(f"Invalid slicing resulting in no data (start={start}, end={stop}, step={step}).")

        if isinstance(task_index, np.ndarray):
            task_index = list(task_index)

        if isinstance(task_index, list):
            task_index = [
                t if t >= 0 else _handle_negative_indexes(t, len(self)) for t in task_index
            ]
            if len(t.shape) == 2:
                indexes = np.unique(np.where(t[:, task_index] == 1)[0])
            else:
                indexes = np.where(np.isin(t, task_index))[0]
        else:
            if task_index < 0:
                task_index = _handle_negative_indexes(task_index, len(self))

            if len(t.shape) == 2:
                indexes = np.where(t[:, task_index] == 1)[0]
            else:
                indexes = np.where(t == task_index)[0]

        selected_x = x[indexes]
        selected_y = y[indexes]
        selected_t = t[indexes]

        if self.cl_dataset.need_class_remapping:  # TODO: to remove with TransformIncremental
            # A remapping of the class ids is done to handle some special cases
            # like PermutedMNIST or RotatedMNIST.
            selected_y = self.cl_dataset.class_remapping(selected_y)

        return selected_x, selected_y, selected_t, task_index


def _handle_negative_indexes(index: int, total_len: int) -> int:
    while index < 0:
        index += total_len
    return index



import statistics

import numpy as np

from continuum.metrics.base_logger import _BaseLogger
from continuum.metrics.utils import require_subset, cache
from continuum.metrics.metrics import accuracy, \
    get_model_size_growth, \
    _get_R_ij, \
    forgetting, \
    accuracy_A, \
    remembering, \
    positive_backward_transfer, \
    forward_transfer, \
    backward_transfer


class Logger(_BaseLogger):
    def __init__(self, list_keywords=["performance"], list_subsets=["train", "test"], root_log=None):
        super().__init__(root_log=root_log, list_keywords=list_keywords, list_subsets=list_subsets)

    def log(self):
        print(f"Task id={self.nb_tasks}, acc={self.accuracy}, avg-acc={self.average_incremental_accuracy}")

    @property
    def nb_tasks(self):
        return self.current_task

    def _conv_list_vector(self, list_vector):
        if len(list_vector) > 1:
            vector = np.concatenate(list_vector)
        else:
            vector = list_vector[0]
        return vector

    def _get_best_epochs_perf(self, subset):
        """If there is no eval data, we assume that the best epoch for each task is the last one"""

        last_epoch_pred = []
        last_epoch_targets = []
        last_epoch_task_ids = []
        for task_id in range(len(self.logger_dict[subset]["performance"])):
            predictions = self.logger_dict[subset]["performance"][task_id][-1]["predictions"]
            targets = self.logger_dict[subset]["performance"][task_id][-1]["targets"]
            task_id = self.logger_dict[subset]["performance"][task_id][-1]["task_ids"]

            last_epoch_pred.append(predictions)
            last_epoch_targets.append(targets)
            last_epoch_task_ids.append(task_id)

        return last_epoch_pred, last_epoch_targets, last_epoch_task_ids

    def _get_best_epochs_data(self, keyword, subset):
        assert keyword != "performance", f"this method is not mode for performance keyword use _get_best_epochs_perf"
        list_values = []
        for task_id in range(self.current_task):
            list_values.append(self.logger_dict[subset][keyword][task_id][-1])
        return list_values

    def _get_best_epochs(self, keyword="performance", subset="train"):
        if keyword == "performance":
            values = self._get_best_epochs_perf(subset)
        else:
            values = self._get_best_epochs_data(keyword, subset)
        return values

    def get_logs(self, keyword, subset):
        return self.logger_dict[subset][keyword]

    @property
    @require_subset("train")
    def online_accuracy(self):
        if self._get_current_predictions("train").size == 0:
            raise Exception(
                "You need to call add([prediction, label, task_id]) in order to compute an online accuracy "
                "(add([prediction, label, None]) also works here, task_id is not needed)."
            )
        predictions = self._get_current_predictions("train")
        targets = self._get_current_targets("train")

        return accuracy(predictions, targets)

    @property
    @require_subset("test")
    def accuracy(self):
        return accuracy(
            self._get_current_predictions("test"),
            self._get_current_targets("test")
        )

    @property
    @require_subset("test")
    def accuracy_per_task(self):
        """Returns all task accuracy individually."""
        all_preds, all_targets, all_tasks = self._get_best_epochs(subset="test")
        last_preds, last_targets, last_tasks = all_preds[-1], all_targets[-1], all_tasks[-1]
        correct_pred = last_preds == last_targets

        acc_per_task = []
        for task_id in np.unique(last_tasks):
            indexes = last_tasks == task_id
            acc_per_task.append(correct_pred[indexes].mean())

        return acc_per_task


    @property
    @require_subset("train")
    def online_cumulative_performance(self):
        """Computes the accuracy of last task on the train set.

        Reference:
        * Online Fast Adaptation and Knowledge Accumulation: a New Approach to Continual Learning
          Caccia et al. NeurIPS 2020
        """
        preds = np.concatenate(
            [dict_epoch['predictions'] for dict_epoch in self.logger_dict["train"]["performance"][self.current_task]]
        )
        targets = np.concatenate(
            [dict_epoch['targets'] for dict_epoch in self.logger_dict["train"]["performance"][self.current_task]]
        )
        return accuracy(preds, targets)

    @property
    @require_subset("test")
    def average_incremental_accuracy(self):
        """Computes the average of the accuracies computed after each task.

        Reference:
        * iCaRL: Incremental Classifier and Representation Learning
          Rebuffi et al. CVPR 2017
        """
        all_preds, all_targets, _ = self._get_best_epochs(subset="test")
        return statistics.mean([
            accuracy(all_preds[t], all_targets[t])
            for t in range(len(all_preds))
        ])

    @property
    @require_subset("test")
    def backward_transfer(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return backward_transfer(all_preds, all_targets, task_ids)

    @property
    @require_subset("test")
    def forward_transfer(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return forward_transfer(all_preds, all_targets, task_ids)

    @property
    @require_subset("test")
    def positive_backward_transfer(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return positive_backward_transfer(all_preds, all_targets, task_ids)

    @property
    @require_subset("test")
    def remembering(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return remembering(all_preds, all_targets, task_ids)

    @property
    @require_subset("test")
    def accuracy_A(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return accuracy_A(all_preds, all_targets, task_ids)

    @property
    @require_subset("test")
    def forgetting(self):
        all_preds, all_targets, task_ids = self._get_best_epochs(subset="test")
        return forgetting(all_preds, all_targets, task_ids)

    @property
    def model_size_growth(self):
        assert "model_size" in self.list_keywords
        sizes = self._get_best_epochs("model_size")
        return get_model_size_growth(sizes)



import pytest
import numpy as np
from torch.utils.data import DataLoader

from continuum import ClassIncremental, InstanceIncremental
from continuum.datasets import (
    CIFAR10, CIFAR100, KMNIST, MNIST, CIFARFellowship, FashionMNIST, Fellowship, MNISTFellowship,
    InMemoryDataset, Fellowship
)


@pytest.fixture
def dataset7c():
    return InMemoryDataset(*gen_dataset(7, 0))


@pytest.fixture
def dataset10c():
    return InMemoryDataset(*gen_dataset(10, 1))


@pytest.fixture
def dataset20c():
    return InMemoryDataset(*gen_dataset(20, 2))


def gen_dataset(nb_classes, pixel_value):
    nb_items_per_class = 5

    x_train = np.ones((nb_items_per_class * nb_classes, 32, 32, 3)) * pixel_value
    y_train = []
    for i in range(nb_classes):
        y_train.append(np.ones(nb_items_per_class, dtype=np.int64) * i)
    y_train = np.concatenate(y_train)

    return (x_train, y_train)


@pytest.mark.parametrize("increment", [1, [7, 10, 20]])
def test_inMemory_updateLabels_Fellowship(increment, dataset7c, dataset10c, dataset20c):
    fellow = Fellowship([dataset7c, dataset10c, dataset20c], update_labels=True)

    x, y, t = fellow.get_data()
    assert len(np.unique(t)) == 3
    assert len(np.unique(y)) == 37

    if isinstance(increment, list):
        continuum = ClassIncremental(fellow, increment=increment)
        assert continuum.nb_classes == 37
        assert continuum.nb_tasks == len(increment)
    else:
        continuum = ClassIncremental(fellow, increment=increment)
        assert continuum.nb_tasks == 37
        assert continuum.nb_classes == 37



@pytest.mark.parametrize("increment", [1, [7, 10, 20]])
def test_inMemory_keepLabels_Fellowship(increment, dataset7c, dataset10c, dataset20c):
    fellow = Fellowship([dataset7c, dataset10c, dataset20c], update_labels=False)

    x, y, t = fellow.get_data()
    assert len(np.unique(t)) == 3
    assert len(np.unique(y)) == 20

    if isinstance(increment, list):
        with pytest.raises(Exception):
            continuum = ClassIncremental(fellow, increment=increment)
    else:
        continuum = ClassIncremental(fellow, increment=increment)
        assert continuum.nb_classes == 20
        assert continuum.nb_tasks == 20


@pytest.mark.parametrize("update_labels,nb_tasks", [
    (True, 0),
    (True, 3),
    (False, 0),
    (False, 3),
])
def test_inMemory_Fellowship(update_labels, nb_tasks, dataset7c, dataset10c, dataset20c):
    fellow = Fellowship([dataset7c, dataset10c, dataset20c], update_labels=update_labels)
    continuum = InstanceIncremental(fellow, nb_tasks=nb_tasks)

    assert continuum.nb_tasks == 3


@pytest.mark.slow
@pytest.mark.parametrize("nb_tasks", [0, 3])
def test_MNIST_Fellowship_Instance_Incremental(nb_tasks, tmpdir):
    dataset = MNISTFellowship(data_path=tmpdir, train=True, download=True)
    dataset.get_data()
    continuum = InstanceIncremental(dataset, nb_tasks=nb_tasks)
    assert len(continuum) == 3


@pytest.mark.slow
def test_MNIST_Fellowship_nb_classes(tmpdir):
    dataset = MNISTFellowship(data_path=tmpdir, train=True, download=True)
    x, y, t = dataset.get_data()
    assert len(np.unique(y)) == 30
    dataset = MNISTFellowship(data_path=tmpdir, train=True, download=True, update_labels=False)
    x, y, t = dataset.get_data()
    assert len(np.unique(y)) == 10


@pytest.mark.slow
def test_MNIST_Fellowship(tmpdir):
    dataset = MNISTFellowship(data_path=tmpdir, train=True, download=True)
    dataset.get_data()
    continuum = ClassIncremental(dataset, increment=10)
    assert len(continuum) == 3


@pytest.mark.slow
def test_CIFAR_Fellowship(tmpdir):
    cl_dataset = CIFARFellowship(data_path=tmpdir, train=True, download=True)
    scenario = ClassIncremental(cl_dataset, increment=10)
    assert len(scenario) == 11


@pytest.mark.slow
@pytest.mark.parametrize(
    "list_datasets,nb_tasks", [
        ([MNIST, FashionMNIST], 2),
        ([KMNIST, MNIST, FashionMNIST], 3),
        ([CIFAR10, CIFAR100], 11),
    ]
)
def test_Fellowship_classes(tmpdir, list_datasets, nb_tasks):
    cl_dataset = Fellowship(
        datasets=[d(data_path=tmpdir, download=True, train=True) for d in list_datasets]
    )
    scenario = ClassIncremental(cl_dataset, increment=10)

    assert len(scenario) == nb_tasks
    for task_id, taskset in enumerate(scenario):
        classes = taskset.get_classes()

        # we check if all classes are here
        assert len(classes) == (classes.max() - classes.min() + 1)


@pytest.mark.slow
@pytest.mark.parametrize("list_datasets", [[MNIST, CIFAR10]])
def test_Fellowship_Dimension_Fail(tmpdir, list_datasets):
    cl_dataset = Fellowship(
        datasets=[d(data_path=tmpdir, download=True, train=True) for d in list_datasets]
    )

    # This does not work since CIFAR10 and MNIST data are not same shape
    with pytest.raises(ValueError):
        continuum = ClassIncremental(cl_dataset, increment=10)



import abc
import os
import torch
import numpy as np


def convert_numpy(tensor):
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.cpu().numpy()
    return tensor


class _BaseLogger(abc.ABC):
    def __init__(self, root_log=None, list_keywords=["performance"], list_subsets=["train", "test"]):
        """
        root_log: folder where logged informations will be saved
        list_keywords: keywords indicating the differentes informations to log, they might be chosen by the user
        or specific for use special features of logger: ex: performance or model_size
         list_subsets: list of data subset with distinguished results: ex ["train", "val", "test"] or [train, "test"]
        """
        self.root_log = root_log
        self.list_keywords = list_keywords
        self.list_subsets = list_subsets

        assert self.list_keywords is not None, f" self.list_keywords should contains at list one keyword"
        assert self.list_subsets is not None, f" self.list_subsets should contains at list one subset"
        assert len(self.list_keywords) >= 1, f" self.list_keywords should contains at list one keyword"
        assert len(self.list_subsets) >= 1, f" self.list_subsets should contains at list one subset"
        assert len(self.list_subsets) == len(set(self.list_subsets)), f" There are duplicate in the subset list"

        self.logger_dict = {}

        # create dict base
        for subset in self.list_subsets:
            self.logger_dict[subset] = {}
            for keyword in self.list_keywords:
                self.logger_dict[subset][keyword] = []

        self.current_task = 0
        self.current_epoch = 0

        # add task and epochs in architecture
        self._update_dict_architecture(update_task=True)

    def add(self, value, keyword="performance", subset="train"):
        assert keyword in self.list_keywords, f"Keyword {keyword} is not declared in list_keywords {self.list_keywords}"
        assert subset in self.list_subsets, f"Field {subset} is not declared in list_keywords {self.list_subsets}"

        if keyword == "performance":
            predictions, targets, task_ids = value
            self._add_perf(predictions, targets, task_ids, subset)
        else:
            self._add_value(value, keyword, subset)

    def _get_current_predictions(self, subset="train"):
        return self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["predictions"]

    def _get_current_targets(self, subset="train"):
        return self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["targets"]

    def _get_current_task_ids(self, subset="train"):
        return self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["task_ids"]

    def _add_value(self, tensor, keyword, subset="train"):
        """Add a tensor in the list of the current epoch (tensor can also be a single value) """

        tensor = convert_numpy(tensor)

        self.logger_dict[subset][keyword][self.current_task][self.current_epoch].append(
            tensor)

    def _add_perf(self, predictions, targets, task_ids=None, subset="train"):
        """Special function for performance, so performance can be logged in one line """
        predictions = convert_numpy(predictions)
        targets = convert_numpy(targets)
        task_ids = convert_numpy(task_ids)

        if not isinstance(predictions, np.ndarray):
            raise TypeError(f"Provide predictions as np.ndarray, not {type(predictions).__name__}.")
        if not isinstance(targets, np.ndarray):
            raise TypeError(f"Provide targets as np.ndarray, not {type(targets).__name__}.")

        assert predictions.size == targets.size, f"{predictions.size} vs {targets.size}"

        predictions = np.concatenate([self._get_current_predictions(subset), predictions])
        self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["predictions"] = predictions

        targets = np.concatenate([self._get_current_targets(subset), targets])
        self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["targets"] = targets

        if task_ids is not None:
            task_ids = np.concatenate([self._get_current_task_ids(subset), task_ids])
            self.logger_dict[subset]["performance"][self.current_task][self.current_epoch]["task_ids"] = task_ids

    def _update_dict_architecture(self, update_task=False):
        for keyword in self.list_keywords:
            for subset in self.list_subsets:
                if update_task:
                    self.logger_dict[subset][keyword].append([])
                if keyword == "performance":
                    self.logger_dict[subset][keyword][self.current_task].append({})
                    self.logger_dict[subset][keyword][self.current_task][self.current_epoch][
                        "predictions"] = np.zeros(0)
                    self.logger_dict[subset][keyword][self.current_task][self.current_epoch]["targets"] = np.zeros(0)
                    self.logger_dict[subset][keyword][self.current_task][self.current_epoch]["task_ids"] = np.zeros(0)
                else:
                    self.logger_dict[subset][keyword][self.current_task].append([])

                assert len(self.logger_dict[subset][keyword]) - 1 == self.current_task, \
                    f"the current task index is {self.current_task} while there are" \
                    f" {len(self.logger_dict[subset][keyword]) - 1} past tasks, there is a mismatch"
                assert len(self.logger_dict[subset][keyword][self.current_task]) - 1 == self.current_epoch, \
                    f"the current epoch index is {self.current_epoch} while there are" \
                    f" {len(self.logger_dict[subset][keyword][self.current_task]) - 1} past epoch, there is a mismatch"

    def print_state(self, keyword, subset):

        print(self.logger_dict)

        print(f"**********************")
        print(f"{keyword} on {subset}")
        print(f"**********************")
        for task in range(self.current_task):
            print(f"**********************")
            print(f"Task: {task}")
            for epoch in range(self.current_epoch + 1):
                print(f"Epoch: {task}")
                print(self.logger_dict[subset][keyword][task][epoch])

    def end_epoch(self):
        self.current_epoch += 1
        self._update_dict_architecture(update_task=False)

    def end_task(self, clean=False):
        if self.root_log is not None and clean:
            self._save()
        self.current_task += 1
        self.current_epoch = 0
        self._update_dict_architecture(update_task=True)

    def _save(self):
        import pickle as pkl
        filename = f"logger_dic_task_{self.current_task}.pkl"
        filename = os.path.join(self.root_log, filename)
        with open(filename, 'wb') as f:
            pkl.dump(self.logger_dict, f, pkl.HIGHEST_PROTOCOL)

        # after saving values we remove them from dictionary to save space and memory
        for subset in self.list_subsets:
            for keyword in self.list_keywords:
                self.logger_dict[subset][keyword][self.current_task] = None



import warnings
from copy import copy
from typing import Callable, List, Union

import numpy as np

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import _BaseScenario


class ClassIncremental(_BaseScenario):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: Each new tasks bring new classes only

    :param cl_dataset: A continual dataset.
    :param nb_tasks: The scenario number of tasks.
    :param increment: Either number of classes per task (e.g. increment=2),
                    or a list specifying for every task the amount of new classes
                     (e.g. increment=[5,1,1,1,1]).
    :param initial_increment: A different task size applied only for the first task.
                              Desactivated if `increment` is a list.
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    :param class_order: An optional custom class order, used for NC.
                        e.g. [0,1,2,3,4,5,6,7,8,9] or [5,2,4,1,8,6,7,9,0,3]
    """

    def __init__(
        self,
        cl_dataset: _ContinuumDataset,
        nb_tasks: int = 0,
        increment: Union[List[int], int] = 0,
        initial_increment: int = 0,
        transformations: Union[List[Callable], List[List[Callable]]] = None,
        class_order: Union[List[int], None]=None
    ) -> None:

        super().__init__(cl_dataset=cl_dataset, nb_tasks=nb_tasks, transformations=transformations)

        self.increment = increment
        self.initial_increment = initial_increment
        self.class_order = class_order

        self._nb_tasks = self._setup(nb_tasks)

    def _setup(self, nb_tasks: int) -> int:

        x, y, _ = self.cl_dataset.get_data()
        unique_classes = np.unique(y)

        if self.class_order is None:
            if self.cl_dataset.class_order is not None:
                self.class_order = self.cl_dataset.class_order
            else:
                self.class_order = unique_classes
        self.class_order = list(self.class_order)

        if len(np.unique(self.class_order)) != len(self.class_order):
            raise ValueError(f"Invalid class order, duplicates found: {self.class_order}.")

        new_y = np.vectorize(self.class_order.index)(y)

        # Increments setup
        self.class_order = np.array(self.class_order)
        if nb_tasks <= 0:
            # The number of tasks is left unspecified, thus it will be determined
            # by the specified increments.
            self.increments = self._define_increments(
                self.increment, self.initial_increment, unique_classes
            )
        else:
            # A fixed number of tasks is required, thus the all increments will
            # be equal among tasks.
            if self.increment > 0:
                warnings.warn(
                    f"When both `nb_tasks` (given value = {nb_tasks}) and "
                    f"`increment` (given value = {self.increment} are both set, "
                    "we only consider the number of tasks. The `increment` "
                    "argument is ignored."
                )
            increment = len(unique_classes) / nb_tasks
            if not increment.is_integer():
                raise Exception(
                    f"Invalid number of tasks ({nb_tasks}) for {len(unique_classes)} classes."
                )
            self.increments = [int(increment) for _ in range(nb_tasks)]

        # compute task label
        task_ids = self._set_task_labels(new_y)

        # Dataset with task label
        self.dataset = (x, new_y, task_ids)  # (data, class label, task label)

        return len(np.unique(task_ids))

    def _set_task_labels(self, y: np.ndarray) -> np.ndarray:
        """For each data point, defines a task associated with the data.

        :param y: label tensor
        :param increments: increments contains information about classes per tasks
        :return: tensor of task label
        """
        t = copy(y)  # task label as same size as y

        for task_index, _ in enumerate(self.increments):
            max_class = sum(self.increments[:task_index + 1])
            min_class = sum(self.increments[:task_index])  # 0 when task_index == 0.

            indexes = np.where(np.logical_and(y >= min_class, y < max_class))[0]
            t[indexes] = task_index
        return t

    def _define_increments(
        self, increment: Union[List[int], int], initial_increment: int, unique_classes: List[int]
    ) -> List[int]:

        if isinstance(increment, list):
            # Check if the total number of classes is compatible
            # with increment list and self.nb_classes
            if not sum(increment) == len(unique_classes):
                raise Exception("The increment list is not compatible with the number of classes")

            increments = increment
        elif isinstance(increment, int) and increment > 0:
            increments = []
            if initial_increment:
                increments.append(initial_increment)

            nb_tasks = (len(unique_classes) - initial_increment) / increment
            if not nb_tasks.is_integer():
                raise Exception(
                    "The tasks won't have an equal number of classes"
                    f" with {len(self.class_order)} and increment {increment}"
                )
            increments.extend([increment for _ in range(int(nb_tasks))])
        else:
            raise TypeError(f"Invalid increment={increment}, it must be an int > 0.")

        return increments

    def get_original_targets(self, targets: np.ndarray) -> np.ndarray:
        """Returns the original targets not changed by the custom class order.

        :param targets: An array of targets, as provided by the task datasets.
        :return: An array of targets, with their original values.
        """
        return self.class_order[targets]

    def _select_data_by_classes(self, min_class_id: int, max_class_id: int):
        """Selects a subset of the whole data for a given set of classes.

        :param min_class_id: The minimum class id.
        :param max_class_id: The maximum class id.
        :return: A tuple of numpy array, the first item being the data and the
                 second the associated targets.
        """
        x_, y_, _ = self.dataset

        indexes = np.where(np.logical_and(y_ >= min_class_id, y_ < max_class_id))[0]
        selected_x = x_[indexes]
        selected_y = y_[indexes]

        if self.cl_dataset.need_class_remapping:
            # A remapping of the class ids is done to handle some special cases
            # like PermutedMNIST or RotatedMNIST.
            selected_y = self.cl_dataset.class_remapping(selected_y)

        return selected_x, selected_y



from typing import Tuple, List

import numpy as np

from continuum.tasks import TaskSet


def split_train_val(dataset: TaskSet, val_split: float = 0.1) -> Tuple[TaskSet, TaskSet]:
    """Split train dataset into two datasets, one for training and one for validation.

    :param dataset: A torch dataset, with .x and .y attributes.
    :param val_split: Percentage to allocate for validation, between [0, 1[.
    :return: A tuple a dataset, respectively for train and validation.
    """
    random_state = np.random.RandomState(seed=1)
    indexes = np.arange(len(dataset))
    random_state.shuffle(indexes)

    train_indexes = indexes[int(val_split * len(indexes)):]
    val_indexes = indexes[:int(val_split * len(indexes))]

    x_train, y_train, t_train = dataset.get_raw_samples(train_indexes)
    train_dataset = TaskSet(x_train, y_train, t_train, trsf=dataset.trsf, data_type=dataset.data_type)

    x_val, y_val, t_val = dataset.get_raw_samples(val_indexes)
    val_dataset = TaskSet(x_val, y_val, t_val, trsf=dataset.trsf, data_type=dataset.data_type)

    return train_dataset, val_dataset


def concat(task_sets: List[TaskSet]) -> TaskSet:
    """Concatenate a dataset A with one or many *other* datasets.

    The transformations will be those of the first dataset.

    :param Tasksets: A list of task sets.
    :return: A concatenated task set.
    """
    x, y, t = [], [], []

    data_type = task_sets[0].data_type

    for task_set in task_sets:
        if task_set.data_type != data_type:
            raise Exception(
                f"Invalid data type {task_set.data_type} != {data_type}"
            )

        x.append(task_set._x)
        y.append(task_set._y)
        t.append(task_set._t)

    return TaskSet(
        np.concatenate(x),
        np.concatenate(y),
        np.concatenate(t),
        trsf=task_sets[0].trsf,
        data_type=data_type
    )



import numpy as np
import pytest
from torch.utils.data import DataLoader

from continuum.tasks import split_train_val
from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental


def gen_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


def gen_tensor_data():
    x_train = np.random.rand(20, 100)
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.rand(20, 100)
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


# this function create data with a mismatch between x and y shape
def gen_bad_data():
    nb_classes = 6
    nb_data_x = 10
    nb_data_y = 100

    x_train = []
    y_train = []
    for i in range(nb_classes):
        x_train.append(np.random.randint(100, size=(nb_data_x, 2, 2, 3)).astype(dtype=np.uint8))
        y_train.append(np.ones(nb_data_y) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)

    return (x_train, y_train.astype(int)), (x_test, y_test.astype(int))


# yapf: disable

@pytest.mark.parametrize("increment,initial_increment,nb_tasks", [
    (2, 0, 5),
    (5, 0, 2),
    (1, 5, 6),
    (2, 4, 4),
    ([5, 1, 1, 3], 0, 4)
])
def test_increments(increment, initial_increment, nb_tasks):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=increment, initial_increment=initial_increment)

    assert scenario.nb_tasks == nb_tasks
    seen_tasks = 0

    for task_id, taskset in enumerate(scenario):
        seen_tasks += 1

        if isinstance(increment, list):
            max_class = sum(increment[:task_id + 1])
            min_class = sum(increment[:task_id])
        elif initial_increment:
            max_class = initial_increment + increment * task_id
            min_class = initial_increment + increment * (task_id - 1) if task_id > 0 else 0
        else:
            max_class = increment * (task_id + 1)
            min_class = increment * task_id

        for _ in DataLoader(taskset):
            pass

        assert np.max(taskset._y) == max_class - 1
        assert np.min(taskset._y) == min_class
    assert seen_tasks == nb_tasks


def test_bad_data():
    train, test = gen_bad_data()
    with pytest.raises(ValueError):
        dummy = InMemoryDataset(*train)


@pytest.mark.parametrize("val_split", [0, 0.1, 0.5, 0.8, 1.0])
def test_split_train_val(val_split):
    train, test = gen_data()
    dummy = InMemoryDataset(*train)
    scenario = ClassIncremental(dummy, increment=5)

    for taskset in scenario:
        train_taskset, val_taskset = split_train_val(taskset, val_split=val_split)
        assert int(val_split * len(taskset)) == len(val_taskset)
        assert len(val_taskset) + len(train_taskset) == len(taskset)


@pytest.mark.parametrize("increment,nb_tasks", [
    (2, 5),
    (5, 2),
    (1, 10)
])
def test_tensor_type(increment, nb_tasks):
    train, test = gen_tensor_data()
    dummy = InMemoryDataset(*train, data_type="tensor")
    scenario = ClassIncremental(dummy, increment=increment)

    taskset = scenario[0]
    for x, y, t in DataLoader(taskset):
        continue

    assert scenario.nb_tasks == nb_tasks

@pytest.mark.parametrize("increment,nb_tasks", [
    (2, 5),
    (5, 2),
    (1, 10)
])
def test_tensor_type_get_samples(increment, nb_tasks):
    train, test = gen_tensor_data()
    dummy = InMemoryDataset(*train, data_type="tensor")
    scenario = ClassIncremental(dummy, increment=increment)

    taskset = scenario[0]
    for x, y, t in DataLoader(taskset):
        continue

    x, y, t = taskset.get_random_samples(5)





def require_subset(subset):
    def wrapper1(func):
        def wrapper2(self):
            if subset not in self.list_subsets:
                raise Exception(
                    f"No {subset} predictions have been logged so far which "
                    f"{func.__name__} rely on!"
                )
            return func(self)
        return wrapper2
    return wrapper1


def cache(func):
    def wrapper(self):
        name = f"__cached_{func.__name__}"
        v = self.__dict__.get(name)
        if v is None:
            v = func(self)
            self.__dict__[name] = v
        return v
    return wrapper



from functools import reduce

import numpy as np


def accuracy(task_preds, task_targets):
    """Computes the accuracy of a given task.

    :param task_preds: Predicted labels.
    :param task_targets: Ground-truth targets.
    :return: a float metric between 0 and 1.
    """

    assert task_preds.size > 0
    assert task_targets.size > 0
    assert task_targets.size == task_preds.size, f"{task_targets.size} vs {task_preds.size}"

    metric = (task_preds == task_targets).mean()
    assert 0. <= metric <= 1.0, metric
    return metric


def accuracy_A(all_preds, all_targets, all_tasks):
    """Accuracy as defined in Diaz-Rodriguez and Lomonaco.

    Note that it is slightly different from the normal accuracy as it considers
    each task accuracy with equal weight, while the normal accuracy considers
    the proportion of all targets.

    Example:
    - Given task 1 with 50,000 images and task 2 with 1,000 images.
    - With normal accuracy, task 1 has more importance in the average accuracy.
    - With this accuracy A, task 1 has as much importance as task 2.

    Reference:
    * Don’t forget, there is more than forgetting: newmetrics for Continual Learning
      Diaz-Rodriguez and Lomonaco et al. NeurIPS Workshop 2018

    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    T = len(all_preds)  # Number of seen tasks so far
    # TODO if we take in account zeroshot, we should take the max of all_tasks?
    A = 0.

    for i in range(T):
        for j in range(i + 1):
            A += _get_R_ij(i, j, all_preds, all_targets, all_tasks)

    metric = A / (T * (T + 1) / 2)
    assert 0. <= metric <= 1.0, metric
    return metric


def backward_transfer(all_preds, all_targets, all_tasks):
    """Measures the influence that learning a task has on the performance on previous tasks.

    Reference:
    * Gradient Episodic Memory for Continual Learning
      Lopez-paz & ranzato, NeurIPS 2017

    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    T = len(all_preds)  # Number of seen tasks so far
    # TODO if we take in account zeroshot, we should take the max of all_tasks?
    if T <= 1:
        return 0.
    bwt = 0.

    for i in range(1, T):
        for j in range(0, i - 1):
            r_ij = _get_R_ij(i, j, all_preds, all_targets, all_tasks)
            r_jj = _get_R_ij(j, j, all_preds, all_targets, all_tasks)

            bwt += (r_ij - r_jj)

    metric = bwt / (T * (T - 1) / 2)
    assert -1. <= metric <= 1.0, metric
    return metric


def positive_backward_transfer(all_preds, all_targets, all_tasks):
    """Computes the the positive gain of Backward transfer.

    Reference:
    * Don’t forget, there is more than forgetting: newmetrics for Continual Learning
      Diaz-Rodriguez and Lomonaco et al. NeurIPS Workshop 2018

    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    bwt = backward_transfer(all_preds, all_targets, all_tasks)
    metric = max(bwt, 0.)
    assert 0. <= metric <= 1.0, metric
    return metric


def remembering(all_preds, all_targets, all_tasks):
    """Computes the forgetting part of Backward transfer.

    Reference:
    * Don’t forget, there is more than forgetting: newmetrics for Continual Learning
      Diaz-Rodriguez and Lomonaco et al. NeurIPS Workshop 2018

    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    bwt = backward_transfer(all_preds, all_targets, all_tasks)
    metric = 1 - abs(min(bwt, 0.))
    assert 0. <= metric <= 1.0, metric
    return metric


def forward_transfer(all_preds, all_targets, all_tasks):
    """Measures the influence that learning a task has on the performance of future tasks.

    Reference:
    * Gradient Episodic Memory for Continual Learning
      Lopez-paz & ranzato, NeurIPS 2017

    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    T = len(all_preds)  # Number of seen tasks so far
    # TODO if we take in account zeroshot, we should take the max of all_tasks?
    if T <= 1:
        return 0.

    fwt = 0.
    for i in range(T):
        for j in range(i):
            fwt += _get_R_ij(i, j, all_preds, all_targets, all_tasks)

    metric = fwt / (T * (T - 1) / 2)
    assert -1. <= metric <= 1.0, metric
    return metric


def forgetting(all_preds, all_targets, all_tasks):
    """Measures the average forgetting.

    Reference:
    * Riemannian Walk for Incremental Learning: Understanding Forgetting and Intransigence
      Chaudhry et al. ECCV 2018

    See eq. 3.
    """
    k = len(all_preds)  # Number of seen tasks so far
    # TODO if we take in account zeroshot, we should take the max of all_tasks?
    if k <= 1:
        return 0.

    f = 0.
    for j in range(k - 2):
        # Accuracy on task j after learning current task k
        a_kj = _get_R_ij(k - 1, j, all_preds, all_targets, all_tasks)
        # Best previous accuracy on task j
        max_a_lj = max(_get_R_ij(l, j, all_preds, all_targets, all_tasks) for l in range(k - 2) if l >= j)
        f += max_a_lj - a_kj  # We want this results to be as low as possible

    metric = f / (k - 1)
    assert -1.0 <= metric <= 1.0, metric
    return metric


def _get_R_ij(i, j, all_preds, all_targets, all_tasks):
    """Computes an accuracy after task i on task j.

    R matrix:

          || T_e1 | T_e2 | T_e3
    ============================|
     T_r1 || R*  | R_ij  | R_ij |
    ----------------------------|
     T_r1 || R*  | R_ij  | R_ij |
    ----------------------------|
     T_r1 || R*  | R_ij  | R_ij |
    ============================|

    R_13 is the R of the first column and the third row.

    From Chaudhry et al., calling this function with (i, j) equals to a_{i,j}.

    Except OOD and Zeroshot, i should be >= j.

    Reference:
    * Don’t forget, there is more than forgetting: newmetrics for Continual Learning
      Diaz-Rodriguez and Lomonaco et al. NeurIPS Workshop 2018

    :param i: Task id after which a model was trained.
    :param j: Task id of the test data.
    :param all_preds: All predicted labels up to now.
    :param all_targets: All targets up to now.
    :param all_tasks: All task ids up to now.
    :return: a float metric between 0 and 1.
    """
    preds = all_preds[i]
    targets = all_targets[i]
    tasks = all_tasks[i]

    indexes = np.where(tasks == j)[0]
    if len(indexes) == 0:
        raise ValueError(
            f"You haven't evaluated on any samples of task {j} after "
            f"training on task {i}, therefore it is impossible to compute some "
            "metrics (e.g. FwT, BwT).\n"
            "Either don't use those metrics or evaluate on all seen tasks. "
            "You can do it one by one (`test_scenario[0], test_scenario[1]`) "
            "or all together (`test_scenario[:task_id + 1]`)."
        )
    return (preds[indexes] == targets[indexes]).mean()


def get_model_size(model):
    """Computes the total number of parameters.

    :param model: A Pytorch's nn.Module model.
    :return: The number of parameters.
    """
    nb_params = 0
    # store original state
    orig_state = model.training
    # we want the number of parameter for inference
    model.eval()
    for w in model.parameters():
        if len(w.shape) > 0:  # Tensor
            nb_params += reduce(lambda a, b: a * b, w.shape)
        else:  # Scalar
            nb_params += 1
    # restore previous state
    model.train(orig_state)
    return nb_params


def get_model_size_growth(model_sizes):
    """Computes the growth of the model sizes.

    Same as model size efficiency but with a less missleading name
    Reference:
    * Don’t forget, there is more than forgetting: newmetrics for Continual Learning
      Diaz-Rodriguez and Lomonaco et al. NeurIPS Workshop 2018

    :param model_sizes: A list of number of parameters, each being computed after a task.
    :return: a float metric between 0 and 1.
    """
    T = len(model_sizes)
    if T <= 1:
        return 1.0

    ms = 0.
    for i in range(T):
        ms += (model_sizes[0][0] / model_sizes[i][-1])

    metric = min(1., ms / T)
    assert 0. <= metric <= 1.0, metric
    return metric



import numpy as np
import pytest

from continuum.datasets import InMemoryDataset
from continuum.scenarios import InstanceIncremental

# yapf: disable


@pytest.fixture
def numpy_data():
    nb_classes = 6
    nb_data = 100

    x_train = []
    y_train = []
    for i in range(nb_classes):
        x_train.append(np.ones((nb_data, 4, 4, 3), dtype=np.uint8) * i)
        y_train.append(np.ones(nb_data) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)

    return (x_train, y_train.astype(int)), (x_test, y_test.astype(int))


@pytest.mark.parametrize("nb_tasks,nb_tasks_gt", [
    (2, 2),
    (6, 6),
    (1, 1),
])
def test_instance_auto_nb_tasks(numpy_data, nb_tasks, nb_tasks_gt):
    """Test the InstanceIncremental loader when the dataset doesn't provide
    any default number of tasks."""
    train, test = numpy_data
    dummy = InMemoryDataset(*train)
    scenario = InstanceIncremental(dummy, nb_tasks=nb_tasks)

    nb_classes = scenario.nb_classes

    assert len(scenario) == nb_tasks_gt
    for task_id, train_dataset in enumerate(scenario):
        assert nb_classes == len(np.unique(train_dataset._y))


@pytest.fixture
def numpy_data_per_task():
    nb_classes = 6
    nb_tasks = 3
    nb_data = 100

    x_train = []
    y_train = []
    t_train = []
    for i in range(nb_tasks):
        for j in range(nb_classes):
            x_train.append(np.ones((nb_data, 4, 4, 3), dtype=np.uint8) * i)
            y_train.append(np.ones(nb_data) * j)
            t_train.append(np.ones(nb_data) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)
    t_train = np.concatenate(t_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)
    t_test = np.copy(t_train)

    return (x_train, y_train.astype(int), t_train), (x_test, y_test.astype(int), t_test)


@pytest.mark.parametrize("nb_tasks,nb_tasks_gt", [
    (2, 2),
    (6, 6),
    (None, 3),
])
def test_instance_default_nb_tasks(numpy_data_per_task, nb_tasks, nb_tasks_gt):
    """Test the InstanceIncremental loader when the dataset does provide
    a default number of tasks."""
    train, test = numpy_data_per_task

    x_train, y_train, t_train = train

    dummy = InMemoryDataset(x_train, y_train, t=t_train)

    scenario = InstanceIncremental(dummy, nb_tasks=nb_tasks)
    nb_classes = scenario.nb_classes

    assert len(scenario) == nb_tasks_gt, (nb_tasks, nb_tasks_gt)
    for task_id, train_dataset in enumerate(scenario):
        assert nb_classes == len(np.unique(train_dataset._y))

        unique_pixels = np.unique(train_dataset._x)
        if nb_tasks is None:
            assert len(unique_pixels) == 1 and unique_pixels[0] == float(task_id)


@pytest.mark.parametrize("nb_tasks,error", [(301, "error"), (300, None), (300, "warning")])
def test_too_many_tasks(numpy_data_per_task, nb_tasks, error):
    train, test = numpy_data_per_task
    x_train, y_train, t_train = train

    if error == "warning":
        x_train = x_train[:-50]
        y_train = y_train[:-50]
        t_train = t_train[:-50]

    dummy = InMemoryDataset(x_train, y_train, t=t_train)

    if error == "error":
        with pytest.raises(Exception):
            scenario = InstanceIncremental(dummy, nb_tasks=nb_tasks)
    elif error == "warning":
        with pytest.warns(Warning):
            scenario = InstanceIncremental(dummy, nb_tasks=nb_tasks)
    else:
        scenario = InstanceIncremental(dummy, nb_tasks=nb_tasks)



@pytest.mark.parametrize("nb_tasks", [None, 0, -1])
def test_invalid(numpy_data_per_task, nb_tasks):
    train, test = numpy_data_per_task

    x_train, y_train, t_train = train
    x_test, y_test, t_test = test

    dummy = InMemoryDataset(x_train, y_train)

    with pytest.raises(Exception):
        InstanceIncremental(dummy, nb_tasks=nb_tasks)


@pytest.fixture
def equal_data():
    x = np.ones((100, 4, 4, 3), dtype=np.uint8)
    y = np.concatenate((
        np.ones((25,), dtype=np.uint32) * 0,
        np.ones((50,), dtype=np.uint32) * 1,
        np.ones((10,), dtype=np.uint32) * 2,
        np.ones((15,), dtype=np.uint32) * 3,
    ))
    return x, y


def test_instance_data_split_equally(equal_data):
    dataset = InMemoryDataset(*equal_data)
    scenario = InstanceIncremental(dataset, nb_tasks=5)

    c = 0

    for taskset in scenario:
        bincount = np.bincount(taskset._y)
        assert bincount[0] == 5
        assert bincount[1] == 10
        assert bincount[2] == 2
        assert bincount[3] == 3

        c += len(taskset)

    assert c == 100


@pytest.fixture
def unequal_data():
    y = np.concatenate((
        np.ones((22,), dtype=np.uint32) * 0,
        np.ones((53,), dtype=np.uint32) * 1,
        np.ones((19,), dtype=np.uint32) * 2,
        np.ones((7,), dtype=np.uint32) * 3,
    ))
    x = np.ones((len(y), 4, 4, 3), dtype=np.uint8)
    return x, y


def test_instance_data_split_not_equally(unequal_data):
    x, y = unequal_data
    dataset = InMemoryDataset(x, y)
    scenario = InstanceIncremental(dataset, nb_tasks=5)

    c_0, c_1, c_2, c_3 = 0, 0, 0, 0

    for taskset in scenario:
        bincount = np.bincount(taskset._y)
        c_0 += bincount[0]
        c_1 += bincount[1]
        c_2 += bincount[2]
        c_3 += bincount[3]

    bincount = np.bincount(y)
    assert c_0 == bincount[0]
    assert c_1 == bincount[1]
    assert c_2 == bincount[2]
    assert c_3 == bincount[3]
    assert c_0 + c_1 + c_2 + c_3 == len(x)
